!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!  Program   : m a k e _ i s m i p _ o u t p u t . F 9 0
!
!> @file
!!
!! Generating ISMIP output from the time-dependent NetCDF data produced by
!! SICOPOLIS.
!!
!! @section Date
!!
!! 2019-01-02
!!
!! @section Copyright
!!
!! Copyright 2010-2019 Ralf Greve
!!
!! @section License
!!
!! This file is part of SICOPOLIS.
!!
!! SICOPOLIS is free software: you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation, either version 3 of the License, or
!! (at your option) any later version.
!!
!! SICOPOLIS is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with SICOPOLIS.  If not, see <http://www.gnu.org/licenses/>.
!<
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!-------- Settings --------

#define OUTPUT_TYPE 1
!                     Type of ISMIP output to be generated:
!                      1 - complete output
!                      2 - only scalar variables

#define RECORD_SAMPLING 10
!                     Sampling of the time-dependent NetCDF data
!                      1 - output made for every record
!                      2 - output made for every 2nd record,
!                          flux variables averaged over 2 records
!                      3 - output made for every 3rd record,
!                          flux variables averaged over 3 records
!                      etc.

#define TIME_UNIT 1
!                     Time unit of ISMIP output to be generated:
!                      1 - seconds (ISMIP default)
!                      2 - years

!-------- Inclusion of specification header --------

#include RUN_SPECS_HEADER

!-------------------------------------------------------------------------------
!> Declarations of kind types.
!<------------------------------------------------------------------------------
module make_ismip_output_types

integer, parameter :: i1b = selected_int_kind(2)   !< 1-byte integers
!!! integer, parameter :: i2b = selected_int_kind(4)   ! 2-byte integers
integer, parameter :: i4b = selected_int_kind(9)   ! 4-byte integers
integer, parameter :: sp  = kind(1.0)              ! single-precision reals
integer, parameter :: dp  = kind(1.0d0)            ! double-precision reals

end module make_ismip_output_types

!-------------------------------------------------------------------------------
!> Main program:
!! Generating ISMIP output from the time-dependent NetCDF data produced by
!! SICOPOLIS.
!<------------------------------------------------------------------------------
program make_ismip_output

use make_ismip_output_types

implicit none

integer(i4b)       :: i, j, n, n1, nn, n_sub, n_sum
integer(i4b)       :: ndat, ncid
integer(i4b)       :: n_help, n_1000s, n_100s, n_10s, n_1s
real(sp)           :: r_n_sum_inv
character(len=256) :: runname
character(len=  4) :: ergnum
character          :: ch_1000s, ch_100s, ch_10s, ch_1s
logical            :: flag_init_output

integer(i1b)       :: crs_val
real(sp)           :: crs_false_E_val, crs_false_N_val
real(sp)           :: crs_latitude_origin_val
real(sp)           :: crs_reference_longitude_val
real(sp)           :: crs_standard_parallel_val
real(sp)           :: time_val
real(sp)           :: x_val(0:IMAX)
real(sp)           :: y_val(0:JMAX)
real(sp)           :: lon_val(0:IMAX,0:JMAX)
real(sp)           :: lat_val(0:IMAX,0:JMAX)
real(sp)           :: lim_val
real(sp)           :: limnsw_val
real(sp)           :: iareag_val
real(sp)           :: iareaf_val
real(sp)           :: tendacabf_val, tendacabf_sum
real(sp)           :: tendlibmassbf_val, tendlibmassbf_sum
real(sp)           :: tendlicalvf_val, tendlicalvf_sum
real(sp)           :: tendligroundf_val, tendligroundf_sum
real(sp)           :: lithk_val(0:IMAX,0:JMAX)
real(sp)           :: orog_val(0:IMAX,0:JMAX)
real(sp)           :: topg_val(0:IMAX,0:JMAX)
real(sp)           :: acabf_val(0:IMAX,0:JMAX), acabf_sum(0:IMAX,0:JMAX)
real(sp)           :: aplacabf_val(0:IMAX,0:JMAX), aplacabf_sum(0:IMAX,0:JMAX)
real(sp)           :: libmassbf_val(0:IMAX,0:JMAX), libmassbf_sum(0:IMAX,0:JMAX)
real(sp)           :: hfgeoubed_val(0:IMAX,0:JMAX), hfgeoubed_sum(0:IMAX,0:JMAX)
real(sp)           :: dlithkdt_val(0:IMAX,0:JMAX)
real(sp)           :: uvelsurf_val(0:IMAX,0:JMAX)
real(sp)           :: vvelsurf_val(0:IMAX,0:JMAX)
real(sp)           :: wvelsurf_val(0:IMAX,0:JMAX)
real(sp)           :: horvelsurf_val(0:IMAX,0:JMAX)
real(sp)           :: uvelbase_val(0:IMAX,0:JMAX)
real(sp)           :: vvelbase_val(0:IMAX,0:JMAX)
real(sp)           :: wvelbase_val(0:IMAX,0:JMAX)
real(sp)           :: horvelbase_val(0:IMAX,0:JMAX)
real(sp)           :: uvelmean_val(0:IMAX,0:JMAX)
real(sp)           :: vvelmean_val(0:IMAX,0:JMAX)
real(sp)           :: horvelmean_val(0:IMAX,0:JMAX)
real(sp)           :: litempsnic_val(0:IMAX,0:JMAX)
real(sp)           :: litempbot_val(0:IMAX,0:JMAX)
real(sp)           :: bwa_val(0:IMAX,0:JMAX)
real(sp)           :: bwp_val(0:IMAX,0:JMAX)
real(sp)           :: strbasemag_val(0:IMAX,0:JMAX)
real(sp)           :: licalvf_val(0:IMAX,0:JMAX), licalvf_sum(0:IMAX,0:JMAX)
real(sp)           :: ligroundf_val(0:IMAX,0:JMAX), ligroundf_sum(0:IMAX,0:JMAX)
real(sp)           :: sftgif_val(0:IMAX,0:JMAX)
real(sp)           :: sftgrf_val(0:IMAX,0:JMAX)
real(sp)           :: sftflf_val(0:IMAX,0:JMAX)
character(len= 64) :: crs_grid_mapping_name_val, crs_ellipsoid_val

runname = RUNNAME

#if (defined(OUTPUT_INIT))

#if (OUTPUT_INIT==0)
   flag_init_output = .false.
#elif (OUTPUT_INIT==1)
   flag_init_output = .true.
#else
   stop ' >>> make_ismip_output: OUTPUT_INIT must be either 0 or 1!'
#endif

#else
   flag_init_output = .true.   ! default for undefined parameter OUTPUT_INIT
#endif

#if (OUTPUT_TYPE==1)

#if (OUTPUT==1 || OUTPUT==3)
   if (flag_init_output) then
      ndat = nint((TIME_END0-TIME_INIT0)/DTIME_OUT0)+1
   else
      ndat = nint((TIME_END0-TIME_INIT0)/DTIME_OUT0)
   end if
#else   /* OUTPUT==2 */
   ndat = N_OUTPUT
#endif

#elif (OUTPUT_TYPE==2)

   if (flag_init_output) then
      ndat = nint((TIME_END0-TIME_INIT0)/DTIME_SER0)+1
   else
      ndat = nint((TIME_END0-TIME_INIT0)/DTIME_SER0)
   end if

#else
   stop ' >>> make_ismip_output: OUTPUT_TYPE must be either 1 or 2!'
#endif

#if (OUTPUT_TYPE==1)
   if (ndat > 9999) stop ' >>> make_ismip_output: Too many records!'
#endif

if (flag_init_output) then
   n1 = 1
else
   n1 = RECORD_SAMPLING
end if

nn = 0

do n=n1, ndat, RECORD_SAMPLING

   nn = nn + 1

   n_sum = 0

   tendacabf_sum     = 0.0_sp
   tendlibmassbf_sum = 0.0_sp
   tendlicalvf_sum   = 0.0_sp
   tendligroundf_sum = 0.0_sp
   acabf_sum         = 0.0_sp
   aplacabf_sum      = 0.0_sp
   libmassbf_sum     = 0.0_sp
   hfgeoubed_sum     = 0.0_sp
   licalvf_sum       = 0.0_sp
   ligroundf_sum     = 0.0_sp

   do n_sub=max((n-RECORD_SAMPLING+1),1), min(n,ndat)

      n_sum = n_sum + 1

      n_help  = n_sub
      n_1000s = n_help/1000
      n_help  = n_help-n_1000s*1000
      n_100s  = n_help/100
      n_help  = n_help-n_100s*100
      n_10s   = n_help/10
      n_help  = n_help-n_10s*10
      n_1s    = n_help

      ch_1000s = char(n_1000s+ichar('0'))
      ch_100s  = char(n_100s +ichar('0'))
      ch_10s   = char(n_10s  +ichar('0'))
      ch_1s    = char(n_1s   +ichar('0'))

      ergnum = ch_1000s//ch_100s//ch_10s//ch_1s

      call read_nc(runname, ergnum, n_sub, &
                   crs_val, crs_grid_mapping_name_val, crs_ellipsoid_val, &
                   crs_false_E_val, crs_false_N_val, crs_latitude_origin_val, &
                   crs_reference_longitude_val, crs_standard_parallel_val, &
                   time_val, x_val, y_val, &
                   lon_val, lat_val, &
                   lim_val, limnsw_val, iareag_val, iareaf_val, &
                   tendacabf_val, tendlibmassbf_val, &
                   tendlicalvf_val, tendligroundf_val, &
                   lithk_val, orog_val, topg_val, &
                   acabf_val, aplacabf_val, libmassbf_val, hfgeoubed_val, &
                   dlithkdt_val, &
                   uvelsurf_val, vvelsurf_val, wvelsurf_val, horvelsurf_val, &
                   uvelbase_val, vvelbase_val, wvelbase_val, horvelbase_val, &
                   uvelmean_val, vvelmean_val, horvelmean_val, &
                   litempsnic_val, litempbot_val, &
                   bwa_val, bwp_val, strbasemag_val, &
                   licalvf_val, ligroundf_val, &
                   sftgif_val, sftgrf_val, sftflf_val)

      tendacabf_sum     = tendacabf_sum     + tendacabf_val
      tendlibmassbf_sum = tendlibmassbf_sum + tendlibmassbf_val
      tendlicalvf_sum   = tendlicalvf_sum   + tendlicalvf_val
      tendligroundf_sum = tendligroundf_sum + tendligroundf_val
      acabf_sum         = acabf_sum         + acabf_val
      aplacabf_sum      = aplacabf_sum      + aplacabf_val
      libmassbf_sum     = libmassbf_sum     + libmassbf_val
      hfgeoubed_sum     = hfgeoubed_sum     + hfgeoubed_val
      licalvf_sum       = licalvf_sum       + licalvf_val
      ligroundf_sum     = ligroundf_sum     + ligroundf_val

   end do

   if (n_sum >= 1) then

      write(6,'(/a,i0,a)') '  * Averaging flux variables over ', &
                                        n_sum, ' record(s) ...'

      r_n_sum_inv = 1.0_sp/real(n_sum,sp)

      tendacabf_val     = tendacabf_sum     * r_n_sum_inv
      tendlibmassbf_val = tendlibmassbf_sum * r_n_sum_inv
      tendlicalvf_val   = tendlicalvf_sum   * r_n_sum_inv
      tendligroundf_val = tendligroundf_sum * r_n_sum_inv
      acabf_val         = acabf_sum         * r_n_sum_inv
      aplacabf_val      = aplacabf_sum      * r_n_sum_inv
      libmassbf_val     = libmassbf_sum     * r_n_sum_inv
      hfgeoubed_val     = hfgeoubed_sum     * r_n_sum_inv
      licalvf_val       = licalvf_sum       * r_n_sum_inv
      ligroundf_val     = ligroundf_sum     * r_n_sum_inv
                          ! averaging of flux variables

   else
      stop ' >>> make_ismip_output: Condition n_sum >= 1 violated!'
   end if

   if (nn==1) then   ! initialisation only once
      call init_ismip_netcdf(runname, &
                    crs_grid_mapping_name_val, crs_ellipsoid_val, &
                    crs_false_E_val, crs_false_N_val, crs_latitude_origin_val, &
                    crs_reference_longitude_val, crs_standard_parallel_val, &
                    ncid)
   end if

   call write_ismip_netcdf(runname, nn, ncid, crs_val, &
                    time_val, x_val, y_val, &
                    lon_val, lat_val, &
                    lim_val, limnsw_val, iareag_val, iareaf_val, &
                    tendacabf_val, tendlibmassbf_val, &
                    tendlicalvf_val, tendligroundf_val, &
                    lithk_val, orog_val, topg_val, &
                    acabf_val, aplacabf_val, libmassbf_val, hfgeoubed_val, &
                    dlithkdt_val, &
                    uvelsurf_val, vvelsurf_val, wvelsurf_val, horvelsurf_val, &
                    uvelbase_val, vvelbase_val, wvelbase_val, horvelbase_val, &
                    uvelmean_val, vvelmean_val, horvelmean_val, &
                    litempsnic_val, litempbot_val, &
                    bwa_val, bwp_val, strbasemag_val, &
                    licalvf_val, ligroundf_val, &
                    sftgif_val, sftgrf_val, sftflf_val)

end do

call close_ismip_netcdf(ncid)

end program make_ismip_output

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!> Reading and processing of data of the time-dependent NetCDF data.
!<++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine read_nc(runname, ergnum, n, &
                   crs_r, crs_grid_mapping_name_r, crs_ellipsoid_r, &
                   crs_false_E_r, crs_false_N_r, crs_latitude_origin_r, &
                   crs_reference_longitude_r, crs_standard_parallel_r, &
                   time_r, x_r, y_r, &
                   lon_r, lat_r, &
                   lim_r, limnsw_r, iareag_r, iareaf_r, &
                   tendacabf_r, tendlibmassbf_r, &
                   tendlicalvf_r, tendligroundf_r, &
                   lithk_r, orog_r, topg_r, &
                   acabf_r, aplacabf_r, libmassbf_r, hfgeoubed_r, &
                   dlithkdt_r, &
                   uvelsurf_r, vvelsurf_r, wvelsurf_r, horvelsurf_r, &
                   uvelbase_r, vvelbase_r, wvelbase_r, horvelbase_r, &
                   uvelmean_r, vvelmean_r, horvelmean_r, &
                   litempsnic_r, litempbot_r, &
                   bwa_r, bwp_r, strbasemag_r, &
                   licalvf_r, ligroundf_r, &
                   sftgif_r, sftgrf_r, sftflf_r)

use make_ismip_output_types
use netcdf

implicit none

character(len=256), intent(in) :: runname
character(len=  4), intent(in) :: ergnum
integer(i4b),       intent(in) :: n

integer(i1b),      intent(out) :: crs_r
real(sp),          intent(out) :: crs_false_E_r, crs_false_N_r
real(sp),          intent(out) :: crs_latitude_origin_r
real(sp),          intent(out) :: crs_reference_longitude_r
real(sp),          intent(out) :: crs_standard_parallel_r
real(sp),          intent(out) :: time_r
real(sp),          intent(out) :: x_r(0:IMAX)
real(sp),          intent(out) :: y_r(0:JMAX)
real(sp),          intent(out) :: lon_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: lat_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: lim_r
real(sp),          intent(out) :: limnsw_r
real(sp),          intent(out) :: iareag_r
real(sp),          intent(out) :: iareaf_r
real(sp),          intent(out) :: tendacabf_r
real(sp),          intent(out) :: tendlibmassbf_r
real(sp),          intent(out) :: tendlicalvf_r
real(sp),          intent(out) :: tendligroundf_r
real(sp),          intent(out) :: lithk_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: orog_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: topg_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: acabf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: aplacabf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: libmassbf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: hfgeoubed_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: dlithkdt_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: uvelsurf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: vvelsurf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: wvelsurf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: horvelsurf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: uvelbase_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: vvelbase_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: wvelbase_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: horvelbase_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: uvelmean_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: vvelmean_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: horvelmean_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: litempsnic_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: litempbot_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: bwa_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: bwp_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: strbasemag_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: licalvf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: ligroundf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: sftgif_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: sftgrf_r(0:IMAX,0:JMAX)
real(sp),          intent(out) :: sftflf_r(0:IMAX,0:JMAX)
character(len=64), intent(out) :: crs_grid_mapping_name_r, crs_ellipsoid_r

integer(i4b)       :: i, j
real(sp)           :: year_to_year_or_sec
character(len=256) :: filename, filename_with_path

integer(i1b), dimension(0:IMAX,0:JMAX) :: maske_erg
real(sp) :: time_erg=0.0_sp, &
            V_tot_erg=0.0_sp, V_af_erg=0.0_sp, &
            A_grounded_erg=0.0_sp, A_floating_erg=0.0_sp, &
            Q_s_erg=0.0_sp, bmb_tot_erg=0.0_sp, calv_tot_erg=0.0_sp, &
            xi_erg(0:IMAX), eta_erg(0:JMAX), &
            sigma_level_c_erg(0:KCMAX), sigma_level_t_erg(0:KTMAX), &
            sigma_level_r_erg(0:KRMAX)
real(sp), dimension(0:IMAX,0:JMAX) :: lon_erg, lat_erg, &
            H_erg, zs_erg, zb_erg, zl_erg, zl0_erg, &
            as_perp_erg, as_perp_apl_erg, Q_bm_erg, Q_tld_erg, q_geo_erg, &
            dH_dtau_erg, &
            vx_s_g_erg, vy_s_g_erg, vz_s_erg, vh_s_erg, &
            vx_b_g_erg, vy_b_g_erg, vz_b_erg, vh_b_erg, &
            vx_m_g_erg, vy_m_g_erg,           vh_m_erg, &
            temp_s_erg, temp_b_erg, temph_b_erg, &
            H_w_erg, p_b_w_erg, &
            tau_b_driving_erg, tau_b_drag_erg, &
            q_gl_g_erg, q_cf_g_erg

integer(i4b) :: ncid, ncv
!     ncid:      ID of the NetCDF file
!     ncv:       Variable ID
integer(i4b) :: nc1cor(1)

real(sp), parameter :: T0  = 273.15_sp
real(sp), parameter :: rho = 910.0_sp

real(sp), parameter :: year_to_sec = 31556926.0_sp

real(sp), parameter :: no_value = -9.999e+03_sp
real(sp), parameter :: eps      =  1.0e-05_sp

!-------- Name of file --------

#if (OUTPUT_TYPE==1)

#if ((OUTPUT==1 || OUTPUT==2) && ERGDAT==0)
   filename = trim(runname)//'_2d_'//trim(ergnum)//'.nc'
#elif ((OUTPUT==1 || OUTPUT==2) && ERGDAT==1)
   filename = trim(runname)//trim(ergnum)//'.nc'
#elif (OUTPUT==3)
   filename = trim(runname)//'_2d_'//trim(ergnum)//'.nc'
#else
   stop ' >>> read_nc: File name cannot be determined!'
#endif

write(6,'(/a)') ' Now reading '//trim(filename)//' ...'

#elif (OUTPUT_TYPE==2)

filename = trim(runname)//'_ser.nc'

write(6,'(/a,i0,a)') ' Now reading '//trim(filename)//', record no. ', n, ' ...'

#endif

!-------- Opening of file --------

filename_with_path = trim(OUTPATH)//'/'//trim(filename)

call check( nf90_open(trim(filename_with_path), NF90_NOWRITE, ncid) )

!-------- Reading of data --------

crs_r = 1                             ! initial value
crs_grid_mapping_name_r = 'unknown'   ! initial value
crs_ellipsoid_r         = 'unknown'   ! initial value
crs_false_E_r             = 0.0_sp    ! initial value
crs_false_N_r             = 0.0_sp    ! initial value
crs_latitude_origin_r     = 0.0_sp    ! initial value
crs_reference_longitude_r = 0.0_sp    ! initial value
crs_standard_parallel_r   = 0.0_sp    ! initial value

#if (OUTPUT_TYPE==1)

if ( nf90_inq_varid(ncid, 'crs', ncv) == nf90_noerr ) then 

   call check( nf90_get_var(ncid, ncv, crs_r) )

   call check( nf90_get_att(ncid, ncv, 'grid_mapping_name', &
                                        crs_grid_mapping_name_r) )
   call check( nf90_get_att(ncid, ncv, 'ellipsoid', &
                                        crs_ellipsoid_r) )
#if (GRID==0 || GRID==1)

   call check( nf90_get_att(ncid, ncv, 'false_easting',  crs_false_E_r) )
   call check( nf90_get_att(ncid, ncv, 'false_northing', crs_false_N_r) )
   call check( nf90_get_att(ncid, ncv, 'latitude_of_projection_origin', &
                                                      crs_latitude_origin_r) )
   call check( nf90_get_att(ncid, ncv, 'straight_vertical_longitude_from_pole', &
                                                  crs_reference_longitude_r) )
   call check( nf90_get_att(ncid, ncv, 'standard_parallel', &
                                                    crs_standard_parallel_r) )
#endif

end if

call check( nf90_inq_varid(ncid, 'time', ncv) )
call check( nf90_get_var(ncid, ncv, time_erg) )

call check( nf90_inq_varid(ncid, 'x', ncv) )
call check( nf90_get_var(ncid, ncv, xi_erg) )

call check( nf90_inq_varid(ncid, 'y', ncv) )
call check( nf90_get_var(ncid, ncv, eta_erg) )

call check( nf90_inq_varid(ncid, 'sigma_level_c', ncv) )
call check( nf90_get_var(ncid, ncv, sigma_level_c_erg) )

call check( nf90_inq_varid(ncid, 'sigma_level_t', ncv) )
call check( nf90_get_var(ncid, ncv, sigma_level_t_erg) )

call check( nf90_inq_varid(ncid, 'sigma_level_r', ncv) )
call check( nf90_get_var(ncid, ncv, sigma_level_r_erg) )

call check( nf90_inq_varid(ncid, 'lon', ncv) )
call check( nf90_get_var(ncid, ncv, lon_erg) )

call check( nf90_inq_varid(ncid, 'lat', ncv) )
call check( nf90_get_var(ncid, ncv, lat_erg) )

call check( nf90_inq_varid(ncid, 'H', ncv) )
call check( nf90_get_var(ncid, ncv, H_erg) )

call check( nf90_inq_varid(ncid, 'zs', ncv) )
call check( nf90_get_var(ncid, ncv, zs_erg) )

call check( nf90_inq_varid(ncid, 'zb', ncv) )
call check( nf90_get_var(ncid, ncv, zb_erg) )

call check( nf90_inq_varid(ncid, 'zl', ncv) )
call check( nf90_get_var(ncid, ncv, zl_erg) )

call check( nf90_inq_varid(ncid, 'zl0', ncv) )
call check( nf90_get_var(ncid, ncv, zl0_erg) )

call check( nf90_inq_varid(ncid, 'as_perp', ncv) )
call check( nf90_get_var(ncid, ncv, as_perp_erg) )

call check( nf90_inq_varid(ncid, 'as_perp_apl', ncv) )
call check( nf90_get_var(ncid, ncv, as_perp_apl_erg) )

call check( nf90_inq_varid(ncid, 'Q_bm', ncv) )
call check( nf90_get_var(ncid, ncv, Q_bm_erg) )

call check( nf90_inq_varid(ncid, 'Q_tld', ncv) )
call check( nf90_get_var(ncid, ncv, Q_tld_erg) )

call check( nf90_inq_varid(ncid, 'q_geo', ncv) )
call check( nf90_get_var(ncid, ncv, q_geo_erg) )

call check( nf90_inq_varid(ncid, 'dH_dt', ncv) )
call check( nf90_get_var(ncid, ncv, dH_dtau_erg) )

call check( nf90_inq_varid(ncid, 'vx_s_g', ncv) )
call check( nf90_get_var(ncid, ncv, vx_s_g_erg) )

call check( nf90_inq_varid(ncid, 'vy_s_g', ncv) )
call check( nf90_get_var(ncid, ncv, vy_s_g_erg) )

call check( nf90_inq_varid(ncid, 'vz_s', ncv) )
call check( nf90_get_var(ncid, ncv, vz_s_erg) )

call check( nf90_inq_varid(ncid, 'vh_s', ncv) )
call check( nf90_get_var(ncid, ncv, vh_s_erg) )

call check( nf90_inq_varid(ncid, 'vx_b_g', ncv) )
call check( nf90_get_var(ncid, ncv, vx_b_g_erg) )

call check( nf90_inq_varid(ncid, 'vy_b_g', ncv) )
call check( nf90_get_var(ncid, ncv, vy_b_g_erg) )

call check( nf90_inq_varid(ncid, 'vz_b', ncv) )
call check( nf90_get_var(ncid, ncv, vz_b_erg) )

call check( nf90_inq_varid(ncid, 'vh_b', ncv) )
call check( nf90_get_var(ncid, ncv, vh_b_erg) )

call check( nf90_inq_varid(ncid, 'vx_m_g', ncv) )
call check( nf90_get_var(ncid, ncv, vx_m_g_erg) )

call check( nf90_inq_varid(ncid, 'vy_m_g', ncv) )
call check( nf90_get_var(ncid, ncv, vy_m_g_erg) )

call check( nf90_inq_varid(ncid, 'vh_m', ncv) )
call check( nf90_get_var(ncid, ncv, vh_m_erg) )

call check( nf90_inq_varid(ncid, 'temp_s', ncv) )
call check( nf90_get_var(ncid, ncv, temp_s_erg) )

call check( nf90_inq_varid(ncid, 'temp_b', ncv) )
call check( nf90_get_var(ncid, ncv, temp_b_erg) )

call check( nf90_inq_varid(ncid, 'temph_b', ncv) )
call check( nf90_get_var(ncid, ncv, temph_b_erg) )

call check( nf90_inq_varid(ncid, 'H_w', ncv) )
call check( nf90_get_var(ncid, ncv, H_w_erg) )

call check( nf90_inq_varid(ncid, 'p_b_w', ncv) )
call check( nf90_get_var(ncid, ncv, p_b_w_erg) )

call check( nf90_inq_varid(ncid, 'tau_b_driving', ncv) )
call check( nf90_get_var(ncid, ncv, tau_b_driving_erg) )

call check( nf90_inq_varid(ncid, 'tau_b_drag', ncv) )
call check( nf90_get_var(ncid, ncv, tau_b_drag_erg) )

call check( nf90_inq_varid(ncid, 'q_gl_g', ncv) )
call check( nf90_get_var(ncid, ncv, q_gl_g_erg) )

call check( nf90_inq_varid(ncid, 'q_cf_g', ncv) )
call check( nf90_get_var(ncid, ncv, q_cf_g_erg) )

call check( nf90_inq_varid(ncid, 'maske', ncv) )
call check( nf90_get_var(ncid, ncv, maske_erg) )

#elif (OUTPUT_TYPE==2)

nc1cor(1) = n

call check( nf90_inq_varid(ncid, 't', ncv) )
call check( nf90_get_var(ncid, ncv, time_erg, start=nc1cor) )

call check( nf90_inq_varid(ncid, 'V_tot', ncv) )
call check( nf90_get_var(ncid, ncv, V_tot_erg, start=nc1cor) )

call check( nf90_inq_varid(ncid, 'V_af', ncv) )
call check( nf90_get_var(ncid, ncv, V_af_erg, start=nc1cor) )

call check( nf90_inq_varid(ncid, 'A_grounded', ncv) )
call check( nf90_get_var(ncid, ncv, A_grounded_erg, start=nc1cor) )

call check( nf90_inq_varid(ncid, 'A_floating', ncv) )
call check( nf90_get_var(ncid, ncv, A_floating_erg, start=nc1cor) )

call check( nf90_inq_varid(ncid, 'Q_s', ncv) )
call check( nf90_get_var(ncid, ncv, Q_s_erg, start=nc1cor) )

call check( nf90_inq_varid(ncid, 'bmb_tot', ncv) )
call check( nf90_get_var(ncid, ncv, bmb_tot_erg, start=nc1cor) )

call check( nf90_inq_varid(ncid, 'calv_tot', ncv) )
call check( nf90_get_var(ncid, ncv, calv_tot_erg, start=nc1cor) )

#endif

!-------- Closing of file --------

call check( nf90_close(ncid) )

!-------- Conversion of read data --------

#if (TIME_UNIT==1)
year_to_year_or_sec = year_to_sec   ! a -> s
#elif (TIME_UNIT==2)
year_to_year_or_sec = 1.0_sp        ! a -> a
#else
stop ' >>> read_nc: TIME_UNIT must be either 1 or 2!'
#endif

time_r           = time_erg * year_to_year_or_sec   ! a -> a | s

#if (OUTPUT_TYPE==1)

lim_r            = no_value
limnsw_r         = no_value
iareag_r         = no_value
iareaf_r         = no_value
tendacabf_r      = no_value
tendlibmassbf_r  = no_value
tendlicalvf_r    = no_value
tendligroundf_r  = no_value

do i=0, IMAX
   x_r(i) = xi_erg(i)   ! m
end do

do j=0, JMAX
   y_r(j) = eta_erg(j)  ! m
end do

sftgif_r = 0.0_sp
sftgrf_r = 0.0_sp
sftflf_r = 0.0_sp

do i=0, IMAX
do j=0, JMAX

   lon_r(i,j)        = lon_erg(i,j)             ! deg E
   lat_r(i,j)        = lat_erg(i,j)             ! deg N
   lithk_r(i,j)      = H_erg(i,j)               ! m
   orog_r(i,j)       = zs_erg(i,j)              ! m
   topg_r(i,j)       = zl_erg(i,j)              ! m
   acabf_r(i,j)      = as_perp_erg(i,j)     * rho/year_to_year_or_sec
                                                ! m/a -> kg/(m2*a) | kg/(m2*s)
   aplacabf_r(i,j)   = as_perp_apl_erg(i,j) * rho/year_to_year_or_sec
                                                ! m/a -> kg/(m2*a) | kg/(m2*s)
   libmassbf_r(i,j)  = -(Q_bm_erg(i,j)+Q_tld_erg(i,j)) * rho/year_to_year_or_sec
                                                ! m/a -> kg/(m2*a) | kg/(m2*s)
   hfgeoubed_r(i,j)  = q_geo_erg(i,j)           ! W/m2
   dlithkdt_r(i,j)   = dH_dtau_erg(i,j) / year_to_year_or_sec   ! m/a | m/s
   uvelsurf_r(i,j)   = vx_s_g_erg(i,j)  / year_to_year_or_sec   ! m/a | m/s
   vvelsurf_r(i,j)   = vy_s_g_erg(i,j)  / year_to_year_or_sec   ! m/a | m/s
   wvelsurf_r(i,j)   = vz_s_erg(i,j)    / year_to_year_or_sec   ! m/a | m/s
   horvelsurf_r(i,j) = vh_s_erg(i,j)    / year_to_year_or_sec   ! m/a | m/s
   uvelbase_r(i,j)   = vx_b_g_erg(i,j)  / year_to_year_or_sec   ! m/a | m/s
   vvelbase_r(i,j)   = vy_b_g_erg(i,j)  / year_to_year_or_sec   ! m/a | m/s
   wvelbase_r(i,j)   = vz_b_erg(i,j)    / year_to_year_or_sec   ! m/a | m/s
   horvelbase_r(i,j) = vh_b_erg(i,j)    / year_to_year_or_sec   ! m/a | m/s
   uvelmean_r(i,j)   = vx_m_g_erg(i,j)  / year_to_year_or_sec   ! m/a | m/s
   vvelmean_r(i,j)   = vy_m_g_erg(i,j)  / year_to_year_or_sec   ! m/a | m/s
   horvelmean_r(i,j) = vh_m_erg(i,j)    / year_to_year_or_sec   ! m/a | m/s
   litempsnic_r(i,j) = temp_s_erg(i,j) + T0     ! C -> K
   litempbot_r(i,j)  = temp_b_erg(i,j) + T0     ! C -> K
   bwa_r(i,j)        = H_w_erg(i,j)             ! m
   bwp_r(i,j)        = p_b_w_erg(i,j)           ! Pa
   strbasemag_r(i,j) = tau_b_drag_erg(i,j)      ! Pa

   if (abs(q_cf_g_erg(i,j)-no_value) > eps) then
      licalvf_r(i,j) = q_cf_g_erg(i,j) * rho/year_to_year_or_sec
                                                ! m/a -> kg/(m2*a) | kg/(m2*s)
   else
      licalvf_r(i,j) = no_value
   end if

   ligroundf_r(i,j)  = no_value
                       !!! q_gl_g_erg(i,j) * rho/year_to_year_or_sec
                                                ! m/a -> kg/(m2*a) | kg/(m2*s)

   if ((maske_erg(i,j)==0_i1b).or.(maske_erg(i,j)==3_i1b)) &
      sftgif_r(i,j)  = 1.0_sp   ! grounded or floating ice

   if (maske_erg(i,j)==0_i1b) &
      sftgrf_r(i,j)  = 1.0_sp   ! grounded ice

   if (maske_erg(i,j)==3_i1b) &
      sftflf_r(i,j)  = 1.0_sp   ! floating ice

end do
end do

#elif (OUTPUT_TYPE==2)

lim_r            = V_tot_erg * rho       ! m3 -> kg
limnsw_r         = V_af_erg  * rho       ! m3 -> kg
iareag_r         = A_grounded_erg        ! m2
iareaf_r         = A_floating_erg        ! m2
tendacabf_r      = Q_s_erg * rho/year_to_year_or_sec
                               ! m3/a -> kg/a | kg/s
tendlibmassbf_r  = bmb_tot_erg * rho/year_to_year_or_sec
                               ! m3/a -> kg/a | kg/s
tendlicalvf_r    = calv_tot_erg * rho/year_to_year_or_sec
                               ! m3/a -> kg/a | kg/s
tendligroundf_r  = no_value
                               ! kg/a | kg/s

x_r          = no_value
y_r          = no_value
lon_r        = no_value
lat_r        = no_value
lithk_r      = no_value
orog_r       = no_value
topg_r       = no_value
acabf_r      = no_value
aplacabf_r   = no_value
libmassbf_r  = no_value
hfgeoubed_r  = no_value
dlithkdt_r   = no_value
uvelsurf_r   = no_value
vvelsurf_r   = no_value
wvelsurf_r   = no_value
horvelsurf_r = no_value
uvelbase_r   = no_value
vvelbase_r   = no_value
wvelbase_r   = no_value
horvelbase_r = no_value
uvelmean_r   = no_value
vvelmean_r   = no_value
horvelmean_r = no_value
litempsnic_r = no_value
litempbot_r  = no_value
bwa_r        = no_value
bwp_r        = no_value
strbasemag_r = no_value
licalvf_r    = no_value
ligroundf_r  = no_value
sftgif_r     = no_value
sftgrf_r     = no_value
sftflf_r     = no_value

#endif

end subroutine read_nc

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!> Initialisation of ISMIP NetCDF file.
!<++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine init_ismip_netcdf(runname, &
                   crs_grid_mapping_name_val, crs_ellipsoid_val, &
                   crs_false_E_val, crs_false_N_val, crs_latitude_origin_val, &
                   crs_reference_longitude_val, crs_standard_parallel_val, &
                   ncid)

use make_ismip_output_types
use netcdf

implicit none

real(sp),           intent(in) :: crs_false_E_val, crs_false_N_val
real(sp),           intent(in) :: crs_latitude_origin_val
real(sp),           intent(in) :: crs_reference_longitude_val
real(sp),           intent(in) :: crs_standard_parallel_val
character(len= 64), intent(in) :: crs_grid_mapping_name_val, crs_ellipsoid_val
character(len=256), intent(in) :: runname

integer(i4b),      intent(out) :: ncid   ! ID of the NetCDF file

integer(i4b) :: ios
integer(i4b) :: n
integer(i4b) :: ncv
!     ncv:       Variable ID
integer(i4b) :: ncd, nc1d, nc2d(2), nc3d(3)
!     ncd:       Dimension ID
!     nc1d:      Dimension of a 1-d array
!     nc2d:      Vector with the dimensions of a 2-d array
!     nc3d:      Vector with the dimensions of a 3-d array
integer(i4b) :: nc3flag(3), nc4flag(4)
!     nc3flag:   Vector with the 3 possible values of a flag variable
!     nc4flag:   Vector with the 4 possible values of a flag variable
integer(i4b) :: nc1cor(1), nc2cor(2), nc3cor(3)
!     nc1cor(1): Corner of a 1-d array
!     nc2cor(2): Corner of a 2-d array
!     nc3cor(3): Corner of a 3-d array
integer(i4b) :: nc1cnt(1), nc2cnt(2), nc3cnt(3)
!     nc1cnt(1): Count of a 1-d array
!     nc2cnt(2): Count of a 2-d array
!     nc3cnt(3): Count of a 3-d array
character(len=  16) :: ch_date, ch_time, ch_zone
character(len= 256) :: filename, filename_with_path, buffer
character(len=8192) :: long_buffer
character           :: ch_time_unit

#if (TIME_UNIT==1)
ch_time_unit = 's'
#elif (TIME_UNIT==2)
ch_time_unit = 'a'
#else
stop ' >>> init_ismip_netcdf: TIME_UNIT must be either 1 or 2!'
#endif

!-------- Open NetCDF file --------

#if (OUTPUT_TYPE==1)
filename = trim(runname)//'_ismip_'//ch_time_unit//'.nc'
#elif (OUTPUT_TYPE==2)
filename = trim(runname)//'_ismip_'//ch_time_unit//'_scalar.nc'
#endif

write(6,'(/a)') ' Now creating new NetCDF file '//trim(filename)//' ...'

filename_with_path = trim(OUTPATH)//'/'//trim(filename)

call check( nf90_create(trim(filename_with_path), NF90_NOCLOBBER, ncid) )

!-------- Global attributes --------

buffer = 'ISMIP output of simulation '//trim(runname)
call check( nf90_put_att(ncid, NF90_GLOBAL, 'title', trim(buffer)) )

buffer = 'Institute of Low Temperature Science, Hokkaido University, '// &
         'Sapporo, Japan'
call check( nf90_put_att(ncid, NF90_GLOBAL, 'institution', trim(buffer)) )

buffer = 'SICOPOLIS Version '//SICO_VERSION
call check( nf90_put_att(ncid, NF90_GLOBAL, 'source', trim(buffer)) )

call date_and_time(ch_date, ch_time, ch_zone)
buffer = ch_date(1:4)//'-'//ch_date(5:6)//'-'//ch_date(7:8)//' '// &
         ch_time(1:2)//':'//ch_time(3:4)//':'//ch_time(5:6)//' '// &
         ch_zone(1:3)//':'//ch_zone(4:5)//' - Data produced'
call check( nf90_put_att(ncid, NF90_GLOBAL, 'history', trim(buffer)) )

buffer = 'http://www.sicopolis.net/'
call check( nf90_put_att(ncid, NF90_GLOBAL, 'references', trim(buffer)) )

filename_with_path = trim(OUTPATH)//'/'//trim(runname)//'_comment.txt'

open(11, iostat=ios, file=trim(filename_with_path), status='old')
if (ios /= 0) then
   write(6,'(a)') ' >>> init_ismip_netcdf - Warning: No comment file found!'
   long_buffer = 'No additional comments.'
else   ! comment file found
   do n=1, 8192
      read(11,'(a)',end=110) buffer
      if (n==1) then
         long_buffer = trim(buffer)
      else
         long_buffer = trim(long_buffer)//' '//trim(buffer)
      end if
      go to 120
      110 exit
      120 continue
   end do
end if
close(11, status='keep')
call check( nf90_put_att(ncid, NF90_GLOBAL, 'comment', trim(long_buffer)) )

buffer = 'CF-1.4'
call check( nf90_put_att(ncid, NF90_GLOBAL, 'Conventions', trim(buffer)) )

!-------- Definition of the dimensions --------

#if (OUTPUT_TYPE==1)
call check( nf90_def_dim(ncid, 'x', IMAX+1,         ncd) )
call check( nf90_def_dim(ncid, 'y', JMAX+1,         ncd) )
#endif
call check( nf90_def_dim(ncid, 't', NF90_UNLIMITED, ncd) )

!-------- Definition of the variables --------

!  ------ Mapping

#if (OUTPUT_TYPE==1)

!    ---- crs

call check( nf90_def_var(ncid, 'crs', NF90_BYTE, ncv) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping_name', &
                                     trim(crs_grid_mapping_name_val)) )
call check( nf90_put_att(ncid, ncv, 'ellipsoid', &
                                     trim(crs_ellipsoid_val)) )
call check( nf90_put_att(ncid, ncv, 'false_easting', crs_false_E_val) )
call check( nf90_put_att(ncid, ncv, 'false_northing', crs_false_N_val) )
call check( nf90_put_att(ncid, ncv, 'latitude_of_projection_origin', &
                                     crs_latitude_origin_val) )
call check( nf90_put_att(ncid, ncv, 'straight_vertical_longitude_from_pole', &
                                     crs_reference_longitude_val) )
call check( nf90_put_att(ncid, ncv, 'standard_parallel', &
                                     crs_standard_parallel_val) )
#endif

!  ------ Time

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'time', NF90_FLOAT, nc1d, ncv) )
buffer = ch_time_unit
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'time'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Time'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!  ------ Projection coordinates

#if (OUTPUT_TYPE==1)

!    ---- x

call check( nf90_inq_dimid(ncid, 'x', nc1d) )
call check( nf90_def_var(ncid, 'x', NF90_FLOAT, nc1d, ncv) )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'projection_x_coordinate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'x-coordinate of the grid point i'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'axis', 'x') )

!    ---- y

call check( nf90_inq_dimid(ncid, 'y', nc1d) )
call check( nf90_def_var(ncid, 'y', NF90_FLOAT, nc1d, ncv) )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'projection_y_coordinate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'y-coordinate of the grid point j'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'axis', 'y') )

!    ---- lon

call check( nf90_inq_dimid(ncid, 'x', nc2d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc2d(2)) )
call check( nf90_def_var(ncid, 'lon', NF90_FLOAT, nc2d, ncv) )
buffer = 'degrees_E'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'longitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Geographical longitude'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- lat

call check( nf90_inq_dimid(ncid, 'x', nc2d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc2d(2)) )
call check( nf90_def_var(ncid, 'lat', NF90_FLOAT, nc2d, ncv) )
buffer = 'degrees_N'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'latitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Geographical latitude'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

#endif

!  ------ Scalar output variables

#if (OUTPUT_TYPE==2)

!    ---- lim

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'lim', NF90_FLOAT, nc1d, ncv) )
buffer = 'kg'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_mass'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Total ice mass'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- limnsw

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'limnsw', NF90_FLOAT, nc1d, ncv) )
buffer = 'kg'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_mass_not_displacing_sea_water'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Mass above floatation'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- iareag

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'iareag', NF90_FLOAT, nc1d, ncv) )
buffer = 'm2'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'grounded_land_ice_area'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Grounded ice area'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- iareaf

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'iareaf', NF90_FLOAT, nc1d, ncv) )
buffer = 'm2'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'floating_ice_shelf_area'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Floating ice area'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- tendacabf

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'tendacabf', NF90_FLOAT, nc1d, ncv) )
buffer = 'kg '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'tendency_of_land_ice_mass_due_to_surface_mass_balance'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Total SMB flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- tendlibmassbf

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'tendlibmassbf', NF90_FLOAT, nc1d, ncv) )
buffer = 'kg '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'tendency_of_land_ice_mass_due_to_basal_mass_balance'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Total BMB flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- tendlicalvf

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'tendlicalvf', NF90_FLOAT, nc1d, ncv) )
buffer = 'kg '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'tendency_of_land_ice_mass_due_to_calving'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Total calving flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

!    ---- tendligroundf

call check( nf90_inq_dimid(ncid, 't', nc1d) )
call check( nf90_def_var(ncid, 'tendligroundf', NF90_FLOAT, nc1d, ncv) )
buffer = 'kg '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'tendency_of_land_ice_mass_due_to_flux_at_grounding_line'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Total grounding line flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )

#endif

!  ------ 2D output variables

#if (OUTPUT_TYPE==1)

!    ---- lithk

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'lithk', NF90_FLOAT, nc3d, ncv) )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Ice thickness'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- orog

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'orog', NF90_FLOAT, nc3d, ncv) )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'surface_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Surface elevation'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- topg

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'topg', NF90_FLOAT, nc3d, ncv) )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'bedrock_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Bedrock elevation'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- acabf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'acabf', NF90_FLOAT, nc3d, ncv) )
buffer = 'kg m-2 '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_surface_specific_mass_balance_flux'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Surface mass balance flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- aplacabf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'aplacabf', NF90_FLOAT, nc3d, ncv) )
buffer = 'kg m-2 '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'applied_land_ice_surface_specific_mass_balance_flux'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Applied surface mass balance flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- libmassbf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'libmassbf', NF90_FLOAT, nc3d, ncv) )
buffer = 'kg m-2 '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_basal_specific_mass_balance_flux'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Basal mass balance flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- hfgeoubed

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'hfgeoubed', NF90_FLOAT, nc3d, ncv) )
buffer = 'W m-2'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'upward_geothermal_heat_flux_at_ground_level'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Geothermal heat flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- dlithkdt

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'dlithkdt', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'tendency_of_land_ice_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Ice thickness imbalance'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- uvelsurf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'uvelsurf', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_surface_x_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Surface velocity in x'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- vvelsurf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'vvelsurf', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_surface_y_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Surface velocity in y'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- wvelsurf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'wvelsurf', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_surface_upward_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Surface velocity in z'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- horvelsurf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'horvelsurf', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_surface_horizontal_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Horizontal surface velocity'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- uvelbase

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'uvelbase', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_basal_x_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Basal velocity in x'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- vvelbase

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'vvelbase', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_basal_y_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Basal velocity in y'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- wvelbase

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'wvelbase', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_basal_upward_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Basal velocity in z'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- horvelbase

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'horvelbase', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_basal_horizontal_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Horizontal basal velocity'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- uvelmean

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'uvelmean', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_vertical_mean_x_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Mean velocity in x'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- vvelmean

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'vvelmean', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_vertical_mean_y_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Mean velocity in y'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- horvelmean

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'horvelmean', NF90_FLOAT, nc3d, ncv) )
buffer = 'm '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_vertical_mean_horizontal_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Horizontal mean velocity'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- litempsnic

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'litempsnic', NF90_FLOAT, nc3d, ncv) )
buffer = 'K'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'temperature_at_ground_level_in_snow_or_firn'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Surface temperature'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- litempbot

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'litempbot', NF90_FLOAT, nc3d, ncv) )
buffer = 'K'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_basal_temperature'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Basal temperature'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- bwa

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'bwa', NF90_FLOAT, nc3d, ncv) )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'subglacial_water_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Effective thickness of subglacial water'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- bwp

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'bwp', NF90_FLOAT, nc3d, ncv) )
buffer = 'Pa'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'basal_water_pressure'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Basal water pressure'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- strbasemag

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'strbasemag', NF90_FLOAT, nc3d, ncv) )
buffer = 'Pa'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'magnitude_of_land_ice_basal_drag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Basal drag'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- licalvf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'licalvf', NF90_FLOAT, nc3d, ncv) )
buffer = 'kg m-2 '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_specific_mass_flux_due_to_calving'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Calving flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- ligroundf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'ligroundf', NF90_FLOAT, nc3d, ncv) )
buffer = 'kg m-2 '//ch_time_unit//'-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_specific_mass_flux_at_grounding_line'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Grounding line flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- sftgif

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'sftgif', NF90_FLOAT, nc3d, ncv) )
buffer = '1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'land_ice_area_fraction'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Land ice area fraction'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- sftgrf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'sftgrf', NF90_FLOAT, nc3d, ncv) )
buffer = '1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'grounded_ice_sheet_area_fraction'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Grounded ice sheet area fraction'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

!    ---- sftflf

call check( nf90_inq_dimid(ncid, 'x', nc3d(1)) )
call check( nf90_inq_dimid(ncid, 'y', nc3d(2)) )
call check( nf90_inq_dimid(ncid, 't', nc3d(3)) )
call check( nf90_def_var(ncid, 'sftflf', NF90_FLOAT, nc3d, ncv) )
buffer = '1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)) )
buffer = 'floating_ice_sheet_area_fraction'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)) )
buffer = 'Floating ice sheet area fraction'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)) )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs') )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon') )

#endif

!    ---- End of the definitions

call check( nf90_enddef(ncid) )

end subroutine init_ismip_netcdf

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!> Writing of ISMIP NetCDF file.
!<++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine write_ismip_netcdf(runname, n, ncid, &
                    crs_val, time_val, x_val, y_val, &
                    lon_val, lat_val, &
                    lim_val, limnsw_val, iareag_val, iareaf_val, &
                    tendacabf_val, tendlibmassbf_val, &
                    tendlicalvf_val, tendligroundf_val, &
                    lithk_val, orog_val, topg_val, &
                    acabf_val, aplacabf_val, libmassbf_val, hfgeoubed_val, &
                    dlithkdt_val, &
                    uvelsurf_val, vvelsurf_val, wvelsurf_val, horvelsurf_val, &
                    uvelbase_val, vvelbase_val, wvelbase_val, horvelbase_val, &
                    uvelmean_val, vvelmean_val, horvelmean_val, &
                    litempsnic_val, litempbot_val, &
                    bwa_val, bwp_val, strbasemag_val, &
                    licalvf_val, ligroundf_val, &
                    sftgif_val, sftgrf_val, sftflf_val)

use make_ismip_output_types
use netcdf

implicit none

character(len=256), intent(in) :: runname
integer(i4b),       intent(in) :: n
integer(i4b),       intent(in) :: ncid   ! ID of the NetCDF file
integer(i1b),       intent(in) :: crs_val(1)
real(sp),           intent(in) :: time_val(1)
real(sp),           intent(in) :: x_val(0:IMAX)
real(sp),           intent(in) :: y_val(0:JMAX)
real(sp),           intent(in) :: lon_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: lat_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: lim_val(1)
real(sp),           intent(in) :: limnsw_val(1)
real(sp),           intent(in) :: iareag_val(1)
real(sp),           intent(in) :: iareaf_val(1)
real(sp),           intent(in) :: tendacabf_val(1)
real(sp),           intent(in) :: tendlibmassbf_val(1)
real(sp),           intent(in) :: tendlicalvf_val(1)
real(sp),           intent(in) :: tendligroundf_val(1)
real(sp),           intent(in) :: lithk_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: orog_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: topg_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: acabf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: aplacabf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: libmassbf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: hfgeoubed_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: dlithkdt_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: uvelsurf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: vvelsurf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: wvelsurf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: horvelsurf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: uvelbase_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: vvelbase_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: wvelbase_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: horvelbase_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: uvelmean_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: vvelmean_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: horvelmean_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: litempsnic_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: litempbot_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: bwa_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: bwp_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: strbasemag_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: licalvf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: ligroundf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: sftgif_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: sftgrf_val(0:IMAX,0:JMAX)
real(sp),           intent(in) :: sftflf_val(0:IMAX,0:JMAX)

integer(i4b) :: i, j
integer(i4b) :: ncv
!     ncv:       Variable ID
integer(i4b) :: ncd, nc1d, nc2d(2), nc3d(3)
!     ncd:       Dimension ID
!     nc1d:      Dimension of a 1-d array
!     nc2d:      Vector with the dimensions of a 2-d array
!     nc3d:      Vector with the dimensions of a 3-d array
integer(i4b) :: nc3flag(3), nc4flag(4)
!     nc3flag:   Vector with the 3 possible values of a flag variable
!     nc4flag:   Vector with the 4 possible values of a flag variable
integer(i4b) :: nc1cor(1), nc2cor(2), nc3cor(3)
!     nc1cor(1): Corner of a 1-d array
!     nc2cor(2): Corner of a 2-d array
!     nc3cor(3): Corner of a 3-d array
integer(i4b) :: nc1cnt(1), nc2cnt(2), nc3cnt(3)
!     nc1cnt(1): Count of a 1-d array
!     nc2cnt(2): Count of a 2-d array
!     nc3cnt(3): Count of a 3-d array
character(len= 16) :: ch_date, ch_time, ch_zone
character(len=256) :: filename, buffer
character          :: ch_empty

!-------- Writing of data on NetCDF file --------

write(6,'(/a)') ' Now writing data in new NetCDF file ...'

!  ------ Mapping variable and projection coordinates

#if (OUTPUT_TYPE==1)

if (n==1) then   ! output only once

   call check( nf90_inq_varid(ncid, 'crs', ncv) )
   nc1cor(1) = 1
   nc1cnt(1) = 1
   call check( nf90_put_var(ncid, ncv, crs_val, start=nc1cor, count=nc1cnt) )

   call check( nf90_inq_varid(ncid, 'x', ncv) )
   nc1cor(1) = 1
   nc1cnt(1) = IMAX + 1
   call check( nf90_put_var(ncid, ncv, x_val, start=nc1cor, count=nc1cnt) )

   call check( nf90_inq_varid(ncid, 'y', ncv) )
   nc1cor(1) = 1
   nc1cnt(1) = JMAX + 1
   call check( nf90_put_var(ncid, ncv, y_val, start=nc1cor, count=nc1cnt) )

   call check( nf90_inq_varid(ncid, 'lon', ncv) )
   nc2cor(1) = 1
   nc2cor(2) = 1
   nc2cnt(1) = IMAX + 1
   nc2cnt(2) = JMAX + 1
   call check( nf90_put_var(ncid, ncv, lon_val, start=nc2cor, count=nc2cnt) )

   call check( nf90_inq_varid(ncid, 'lat', ncv) )
   nc2cor(1) = 1
   nc2cor(2) = 1
   nc2cnt(1) = IMAX + 1
   nc2cnt(2) = JMAX + 1
   call check( nf90_put_var(ncid, ncv, lat_val, start=nc2cor, count=nc2cnt) )

end if

#endif

!  ------ Time

call check( nf90_inq_varid(ncid, 'time', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, time_val, start=nc1cor, count=nc1cnt) )

!  ------ Scalar output variables

#if (OUTPUT_TYPE==2)

call check( nf90_inq_varid(ncid, 'lim', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, lim_val, start=nc1cor, count=nc1cnt) )

call check( nf90_inq_varid(ncid, 'limnsw', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, limnsw_val, start=nc1cor, count=nc1cnt) )

call check( nf90_inq_varid(ncid, 'iareag', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, iareag_val, start=nc1cor, count=nc1cnt) )

call check( nf90_inq_varid(ncid, 'iareaf', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, iareaf_val, start=nc1cor, count=nc1cnt) )

call check( nf90_inq_varid(ncid, 'tendacabf', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, tendacabf_val, start=nc1cor, count=nc1cnt) )

call check( nf90_inq_varid(ncid, 'tendlibmassbf', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, tendlibmassbf_val, start=nc1cor, count=nc1cnt) )

call check( nf90_inq_varid(ncid, 'tendlicalvf', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, tendlicalvf_val, start=nc1cor, count=nc1cnt) )

call check( nf90_inq_varid(ncid, 'tendligroundf', ncv) )
nc1cor(1) = n
nc1cnt(1) = 1
call check( nf90_put_var(ncid, ncv, tendligroundf_val, start=nc1cor, count=nc1cnt) )

#endif

!  ------ 2D output variables

#if (OUTPUT_TYPE==1)

call check( nf90_inq_varid(ncid, 'lithk', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, lithk_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'orog', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, orog_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'topg', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, topg_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'acabf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, acabf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'aplacabf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, aplacabf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'libmassbf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, libmassbf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'hfgeoubed', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, hfgeoubed_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'dlithkdt', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, dlithkdt_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'uvelsurf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, uvelsurf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'vvelsurf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, vvelsurf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'wvelsurf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, wvelsurf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'horvelsurf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, horvelsurf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'uvelbase', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, uvelbase_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'vvelbase', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, vvelbase_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'wvelbase', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, wvelbase_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'horvelbase', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, horvelbase_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'uvelmean', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, uvelmean_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'vvelmean', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, vvelmean_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'horvelmean', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, horvelmean_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'litempsnic', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, litempsnic_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'litempbot', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, litempbot_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'bwa', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, bwa_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'bwp', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, bwp_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'strbasemag', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, strbasemag_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'licalvf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, licalvf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'ligroundf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, ligroundf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'sftgif', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, sftgif_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'sftgrf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, sftgrf_val, start=nc3cor, count=nc3cnt) )

call check( nf90_inq_varid(ncid, 'sftflf', ncv) )
nc3cor(1) = 1
nc3cor(2) = 1
nc3cor(3) = n
nc3cnt(1) = IMAX + 1
nc3cnt(2) = JMAX + 1
nc3cnt(3) = 1
call check( nf90_put_var(ncid, ncv, sftflf_val, start=nc3cor, count=nc3cnt) )

#endif

end subroutine write_ismip_netcdf

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!> Closing of ISMIP NetCDF file.
!<++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine close_ismip_netcdf(ncid)

use make_ismip_output_types
use netcdf

implicit none

integer(i4b), intent(in) :: ncid   ! ID of the NetCDF file

write(6,'(/a)') ' Now closing new NetCDF file ...'

call check( nf90_sync(ncid) )

call check( nf90_close(ncid) )

write(6,'(/a/)') ' Done!'

end subroutine close_ismip_netcdf

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!> NetCDF error capturing.
!<++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
subroutine check(status)

use make_ismip_output_types
use netcdf

implicit none

integer(i4b), intent (in) :: status

if (status /= nf90_noerr) then 
   write(6,'(1x,a)') trim(nf90_strerror(status))
   stop ' >>> check: Stopped due to NetCDF error!'
end if

end subroutine check  

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!                        End of make_ismip_output.F90
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
