function [time, x, y, ...
          lithk, orog, topg, ...
          acabf, aplacabf, libmassbf, hfgeoubed, ...
          dlithkdt, ...
          uvelsurf, vvelsurf, wvelsurf, horvelsurf, ...
          uvelbase, vvelbase, wvelbase, horvelbase, ...
          uvelmean, vvelmean, horvelmean, ...
          litempsnic, litempbot, ...
          bwa, bwp, strbasemag, ...
          licalvf, ligroundf, ...
          sftgif, sftgrf, sftflf] ...
        = read_ismip_output(runname, outpath, filename)
% READ_ISMIP_OUTPUT Reading of ISMIP output data.
% =========================================================================
% read_ismip_output  Date: 2017-04-03
%
% Usage: 
%
%   [time, x, y, ...
%    lithk, orog, topg, ...
%    acabf, aplacabf, libmassbf, hfgeoubed, ...
%    dlithkdt, ...
%    uvelsurf, vvelsurf, wvelsurf, horvelsurf, ...
%    uvelbase, vvelbase, wvelbase, horvelbase, ...
%    uvelmean, vvelmean, horvelmean, ...
%    litempsnic, litempbot, ...
%    bwa, bwp, strbasemag, ...
%    licalvf, ligroundf, ...
%    sftgif, sftgrf, sftflf] ...
%      = read_ismip_output(runname, outpath, filename)
%
% Description:
%   read_ismip_output reads NetCDF output data of SICOPOLIS in the
%   ISMIP standard format (as produced by the tool make_ismip_output).
%   Results are returned in the vectors
%   time, x, y, and in the 3-D arrays
%   lithk, orog, topg, acabf, aplacabf, libmassbf, hfgeoubed, dlithkdt,
%   uvelsurf, vvelsurf, wvelsurf, horvelsurf,
%   uvelbase, vvelbase, wvelbase, horvelbase,
%   uvelmean, vvelmean, horvelmean,
%   litempsnic, litempbot, bwa, bwp, strbasemag,
%   licalvf, ligroundf, sftgif, sftgrf, sftflf.
%
% Input:
%   runname  - name of SICOPOLIS simulation (string)
%   outpath  - path of ISMIP output data file (without trailing "/")
%   filename - name of ISMIP output data file
%
% Output:
%   time          - Time [s | a]
%   x             - x-coordinate of the grid point i [m]
%   y             - y-coordinate of the grid point j [m]
%   lithk         - Ice thickness [m]
%   orog          - Surface elevation [m]
%   topg          - Bedrock elevation [m]
%   acabf         - Surface mass balance flux [kg m-2 s-1 | kg m-2 a-1]
%   aplacabf      - Applied surface mass balance flux
%                   [kg m-2 s-1 | kg m-2 a-1]
%   libmassbf     - Basal mass balance flux [kg m-2 s-1 | kg m-2 a-1]
%   hfgeoubed     - Geothermal heat flux [W m-2]
%   dlithkdt      - Ice thickness imbalance [m s-1 | m a-1]
%   uvelsurf      - Surface velocity in x [m s-1 | m a-1]
%   vvelsurf      - Surface velocity in y [m s-1 | m a-1]
%   wvelsurf      - Surface velocity in z [m s-1 | m a-1]
%   horvelsurf    - Horizontal surface velocity [m s-1 | m a-1]
%   uvelbase      - Basal velocity in x [m s-1 | m a-1]
%   vvelbase      - Basal velocity in y [m s-1 | m a-1]
%   wvelbase      - Basal velocity in z [m s-1 | m a-1]
%   horvelbase    - Horizontal basal velocity [m s-1 | m a-1]
%   uvelmean      - Mean velocity in x [m s-1 | m a-1]
%   vvelmean      - Mean velocity in y [m s-1 | m a-1]
%   horvelmean    - Horizontal mean velocity [m s-1 | m a-1]
%   litempsnic    - Surface temperature [K]
%   litempbot     - Basal temperature [K]
%   bwa           - Effective thickness of subglacial water [m]
%   bwp           - Basal water pressure [Pa]
%   strbasemag    - Basal drag [Pa]
%   licalvf       - Calving flux [kg m-2 s-1 | kg m-2 a-1]
%   ligroundf     - Grounding line flux [kg m-2 s-1 | kg m-2 a-1]
%   sftgif        - Land ice area fraction [1]
%   sftgrf        - Grounded ice sheet area fraction [1]
%   sftflf        - Floating ice sheet area fraction [1]
%
% Author:
%   Ralf Greve
% =========================================================================

if nargout ~= 32
   disp('read_ismip_output:')
   disp('Wrong number of output arguments specified in call!')
   return
end

%-------- Opening NetCDF file --------

if strcmp(filename(end-2:end), '.nc')
   disp(' ')
   disp(['Name of ISMIP file: ' filename])
else
   filename = [filename '.nc'];
   disp(' ')
   disp(['Name of ISMIP file: ' filename])
end

try
   ncid = netcdf.open([outpath '/' filename], 'nowrite');
catch
   disp(' ')
   disp(['ISMIP file not found in directory ' outpath '.'])
   outpath = [outpath '/' runname];
   disp(['Trying ' outpath ' instead...'])
   try
      ncid = netcdf.open([outpath '/' filename], 'nowrite');
      disp('ISMIP file found. Continuing...')
   catch err
      disp(['ISMIP file not found in directory ' outpath ' either.'])
      disp('Giving up.')
      rethrow(err)
   end
end

%-------- Reading of data from NetCDF file --------

%  ------ Time

try
   varid = netcdf.inqVarID(ncid,'time');
catch
   varid = netcdf.inqVarID(ncid,'t');
end

time = netcdf.getVar(ncid,varid);

%  ------ Projection coordinates

varid = netcdf.inqVarID(ncid,'x');
x     = netcdf.getVar(ncid,varid);

varid = netcdf.inqVarID(ncid,'y');
y     = netcdf.getVar(ncid,varid);

%  ------ 2D fields

varid       = netcdf.inqVarID(ncid,'lithk');
lithk       = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'orog');
orog        = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'topg');
topg        = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'acabf');
acabf       = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'aplacabf');
aplacabf    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'libmassbf');
libmassbf   = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'hfgeoubed');
hfgeoubed   = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'dlithkdt');
dlithkdt    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'uvelsurf');
uvelsurf    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'vvelsurf');
vvelsurf    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'wvelsurf');
wvelsurf    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'horvelsurf');
horvelsurf  = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'uvelbase');
uvelbase    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'vvelbase');
vvelbase    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'wvelbase');
wvelbase    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'horvelbase');
horvelbase  = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'uvelmean');
uvelmean    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'vvelmean');
vvelmean    = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'horvelmean');
horvelmean  = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'litempsnic');
litempsnic  = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'litempbot');
litempbot   = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'bwa');
bwa         = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'bwp');
bwp         = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'strbasemag');
strbasemag  = netcdf.getVar(ncid,varid);

try
   varid   = netcdf.inqVarID(ncid,'licalvf');
   licalvf = netcdf.getVar(ncid,varid);
catch
   disp(' ')
   disp('read_ismip_output:')
   disp('Variable licalvf not available in read file.')
   licalvf = NaN(size(lithk),'single') ;
end

try
   varid     = netcdf.inqVarID(ncid,'ligroundf');
   ligroundf = netcdf.getVar(ncid,varid);
catch
   disp(' ')
   disp('read_ismip_output:')
   disp('Variable ligroundf not available in read file.')
   ligroundf = NaN(size(lithk),'single') ;
end

varid       = netcdf.inqVarID(ncid,'sftgif');
sftgif      = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'sftgrf');
sftgrf      = netcdf.getVar(ncid,varid);

varid       = netcdf.inqVarID(ncid,'sftflf');
sftflf      = netcdf.getVar(ncid,varid);

%-------- Closing NetCDF file --------

netcdf.close(ncid);

disp(' ')
disp('Done.')
disp(' ')

end % function read_ismip_output
%
