function [time, lim, limnsw, iareag, iareaf, ...
          tendacabf, tendlibmassbf, tendlicalvf, tendligroundf] ...
        = read_ismip_output_scalar(runname, outpath, filename)
% READ_ISMIP_OUTPUT_SCALAR Reading of scalar ISMIP output data.
% =========================================================================
% read_ismip_output_scalar  Date: 2017-04-03
%
% Usage: 
%
%   [time, lim, limnsw, iareag, iareaf, ...
%    tendacabf, tendlibmassbf, tendlicalvf, tendligroundf] ...
%      = read_ismip_output_scalar(runname, outpath, filename)
%
% Description:
%   read_ismip_output_scalar reads scalar NetCDF output data
%   of SICOPOLIS in the ISMIP standard format
%   (as produced by the tool make_ismip_output).
%   Results are returned in the vectors
%   time, lim, limnsw, iareag, iareaf,
%   tendacabf, tendlibmassbf, tendlicalvf, tendligroundf.
%
% Input:
%   runname  - name of SICOPOLIS simulation (string)
%   outpath  - path of ISMIP output data file (without trailing "/")
%   filename - name of ISMIP output data file
%
% Output:
%   time          - Time [s | a]
%   lim           - Total ice mass [kg]
%   limnsw        - Mass above floatation [kg]
%   iareag        - Grounded ice area [m2]
%   iareaf        - Floating ice area [m2]
%   tendacabf     - Total SMB flux [kg s-1 | kg a-1]
%   tendlibmassbf - Total BMB flux [kg s-1 | kg a-1]
%   tendlicalvf   - Total calving flux [kg s-1 | kg a-1]
%   tendligroundf - Total grounding line flux [kg s-1 | kg a-1]
%
% Author:
%   Ralf Greve
% =========================================================================

if nargout ~= 9
   disp('read_ismip_output_scalar:')
   disp('Wrong number of output arguments specified in call!')
   return
end

%-------- Opening NetCDF file --------

if strcmp(filename(end-2:end), '.nc')
   disp(' ')
   disp(['Name of ISMIP file: ' filename])
else
   filename = [filename '.nc'];
   disp(' ')
   disp(['Name of ISMIP file: ' filename])
end

try
   ncid = netcdf.open([outpath '/' filename], 'nowrite');
catch
   disp(' ')
   disp(['ISMIP file not found in directory ' outpath '.'])
   outpath = [outpath '/' runname];
   disp(['Trying ' outpath ' instead...'])
   try
      ncid = netcdf.open([outpath '/' filename], 'nowrite');
      disp('ISMIP file found. Continuing...')
   catch err
      disp(['ISMIP file not found in directory ' outpath ' either.'])
      disp('Giving up.')
      rethrow(err)
   end
end

%-------- Reading of data from NetCDF file --------

%  ------ Time

try
   varid = netcdf.inqVarID(ncid,'time');
catch
   varid = netcdf.inqVarID(ncid,'t');
end

time = netcdf.getVar(ncid,varid);

%  ------ Scalar variables

varid         = netcdf.inqVarID(ncid,'lim');
lim           = netcdf.getVar(ncid,varid);

varid         = netcdf.inqVarID(ncid,'limnsw');
limnsw        = netcdf.getVar(ncid,varid);

varid         = netcdf.inqVarID(ncid,'iareag');
iareag        = netcdf.getVar(ncid,varid);

varid         = netcdf.inqVarID(ncid,'iareaf');
iareaf        = netcdf.getVar(ncid,varid);

varid         = netcdf.inqVarID(ncid,'tendacabf');
tendacabf     = netcdf.getVar(ncid,varid);

varid         = netcdf.inqVarID(ncid,'tendlibmassbf');
tendlibmassbf = netcdf.getVar(ncid,varid);

try
   varid         = netcdf.inqVarID(ncid,'tendlicalvf');
   tendlicalvf   = netcdf.getVar(ncid,varid);
catch
   disp(' ')
   disp('read_ismip_output_scalar:')
   disp('Variable tendlicalvf not available in read file.')
   tendlicalvf = NaN(size(lim),'single') ;
end

try
   varid         = netcdf.inqVarID(ncid,'tendligroundf');
   tendligroundf = netcdf.getVar(ncid,varid);
catch
   disp(' ')
   disp('read_ismip_output_scalar:')
   disp('Variable tendligroundf not available in read file.')
   tendligroundf = NaN(size(lim),'single') ;
end

%-------- Closing NetCDF file --------

netcdf.close(ncid);

disp(' ')
disp('Done.')
disp(' ')

end % function read_ismip_output_scalar
%
