!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
!
!  Module :  o u t p u t _ m
!
!> @file
!!
!! Writing of output data on files.
!!
!! @section Copyright
!!
!! Copyright 2009-2019 Ralf Greve, Reinhard Calov, Thomas Goelles,
!!                     Thorben Dunse
!!
!! @section License
!!
!! This file is part of SICOPOLIS.
!!
!! SICOPOLIS is free software: you can redistribute it and/or modify
!! it under the terms of the GNU General Public License as published by
!! the Free Software Foundation, either version 3 of the License, or
!! (at your option) any later version.
!!
!! SICOPOLIS is distributed in the hope that it will be useful,
!! but WITHOUT ANY WARRANTY; without even the implied warranty of
!! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!! GNU General Public License for more details.
!!
!! You should have received a copy of the GNU General Public License
!! along with SICOPOLIS.  If not, see <http://www.gnu.org/licenses/>.
!<
!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

!-------------------------------------------------------------------------------
!> Writing of output data on files.
!<------------------------------------------------------------------------------
module output_m

  use sico_types_m
  use sico_variables_m
  use sico_vars_m
  use error_m

  implicit none

  real(dp), parameter :: sec_to_year = 1.0_dp/YEAR_SEC

  private
  public :: output1, output2, output4, borehole
#if (defined(ASF))
  public :: output5
#endif

contains

!-------------------------------------------------------------------------------
!> Writing of time-slice files in native binary or NetCDF format.
!<------------------------------------------------------------------------------
subroutine output1(runname, time, delta_ts, glac_index, z_sl, &
                   flag_3d_output, ndat2d, ndat3d)

#if (CALCMOD==1 || CALCMOD==0 || CALCMOD==-1)
  use enth_temp_omega_m, only : enth_fct_temp_omega
#endif

#if (NETCDF==2)   /* time-slice file in NetCDF format */
  use netcdf
  use nc_check_m
#endif

#if (DISC>0)
  use discharge_workers_m, only: dis_perp, cst_dist, cos_grad_tc, mask_mar
#endif

implicit none

real(dp),           intent(in) :: time, delta_ts, glac_index, z_sl
character(len=100), intent(in) :: runname
logical,            intent(in) :: flag_3d_output

integer(i4b),    intent(inout) :: ndat2d, ndat3d

integer(i4b) :: i, j, kc, kt, kr
integer(i4b) :: ndat, ndat_help, ndat_1000s, ndat_100s, ndat_10s, ndat_1s
real(dp), dimension(0:JMAX,0:IMAX) :: H, H_cold, H_temp, dH_dtau
real(dp), dimension(0:JMAX,0:IMAX) :: vx_m_g, vy_m_g
real(dp), dimension(0:JMAX,0:IMAX) :: tau_b_driving, tau_b_drag
real(dp) :: V_tot, V_grounded, V_floating, V_gr_redu, V_af
real(dp) :: A_grounded, A_floating
real(sp) :: lon0, lat0
real(dp) :: rhosw_rho_ratio
character(len=256) :: filename, filename_with_path
character :: ch_1000s, ch_100s, ch_10s, ch_1s
character(len=16) :: ch_date, ch_time, ch_zone

real(dp), parameter :: one_year = 1.0_dp

integer(i1b), dimension(0:IMAX,0:JMAX) :: maske_conv, maske_old_conv, &
                                          mask_ablation_type_conv, &
                                          n_cts_conv
integer(i4b), dimension(0:IMAX,0:JMAX) :: kc_cts_conv
integer(i1b), dimension(0:IMAX,0:JMAX) :: mask_mar_conv
integer(i1b), dimension(0:IMAX,0:JMAX) :: flag_shelfy_stream_x_conv, &
                                          flag_shelfy_stream_y_conv, &
                                          flag_shelfy_stream_conv
integer(i1b), dimension(0:IMAX,0:JMAX) :: flag_grounding_line_1_conv, &
                                          flag_grounding_line_2_conv, &
                                          flag_calving_front_1_conv, &
                                          flag_calving_front_2_conv, &
                                          flag_grounded_front_a_1_conv, &
                                          flag_grounded_front_a_2_conv, &
                                          flag_grounded_front_b_1_conv, &
                                          flag_grounded_front_b_2_conv

real(sp) :: time_conv, delta_ts_conv, glac_index_conv, z_sl_conv, &
            V_tot_conv, V_af_conv, A_grounded_conv, A_floating_conv, &
            H_R_conv, &
            xi_conv(0:IMAX), eta_conv(0:JMAX), &
            sigma_level_c_conv(0:KCMAX), sigma_level_t_conv(0:KTMAX), &
            sigma_level_r_conv(0:KRMAX)
real(sp), dimension(0:IMAX,0:JMAX) :: lambda_conv, phi_conv, &
            lon_conv, lat_conv, &
            temp_s_conv, accum_conv, & 
            snowfall_conv, rainfall_conv, pdd_conv, & 
            as_perp_conv, as_perp_apl_conv, smb_corr_conv, &
            q_geo_conv, &
            zs_conv, zm_conv, zb_conv, zl_conv, zl0_conv, &
            H_cold_conv, H_temp_conv, H_conv, &
            Q_bm_conv, Q_tld_conv, &
            am_perp_conv, &
            qx_conv, qy_conv, &
            dzs_dtau_conv, dzm_dtau_conv, dzb_dtau_conv, dzl_dtau_conv, &
            dH_c_dtau_conv, dH_t_dtau_conv, dH_dtau_conv, &
            vx_b_g_conv, vy_b_g_conv, vz_b_conv, vh_b_conv, &
            vx_s_g_conv, vy_s_g_conv, vz_s_conv, vh_s_conv, &
            vx_m_g_conv, vy_m_g_conv,            vh_m_conv, &
            temp_b_conv, temph_b_conv, &
            tau_b_driving_conv, tau_b_drag_conv, &
            p_b_w_conv, q_w_conv, q_w_x_conv, q_w_y_conv, H_w_conv, &
            q_gl_g_conv, q_cf_g_conv, &
            cst_dist_conv, cos_grad_tc_conv, dis_perp_conv, &
            ratio_sl_x_conv, ratio_sl_y_conv, &
            vis_ave_g_conv, vis_int_g_conv
            
real(sp), dimension(0:IMAX,0:JMAX,0:KCMAX) :: vx_c_conv, vy_c_conv, vz_c_conv, &
                                              temp_c_conv, age_c_conv, &
                                              enth_c_conv, omega_c_conv, &
                                              enh_c_conv
real(sp), dimension(0:IMAX,0:JMAX,0:KTMAX) :: vx_t_conv, vy_t_conv, vz_t_conv, &
                                              omega_t_conv, age_t_conv, &
                                              enth_t_conv, &
                                              enh_t_conv
real(sp), dimension(0:IMAX,0:JMAX,0:KRMAX) :: temp_r_conv

#if (NETCDF==1)   /* time-slice file in native binary format */

integer(i4b) :: ios
character(len=256) :: ch_attr_title, ch_attr_institution, ch_attr_source, &
                      ch_attr_history, ch_attr_references
character(len= 16), parameter :: filename_extension = '.erg'

#elif (NETCDF==2) /* time-slice file in NetCDF format */

integer(i4b) :: ncid, ncv
!     ncid:      ID of the output file
!     ncv:       Variable ID
integer(i4b) :: ncd, nc1d, nc2d(2), nc3d(3)
!     ncd:       Dimension ID
!     nc1d:      Dimension of a 1-d array
!     nc2d:      Vector with the dimensions of a 2-d array
!     nc3d:      Vector with the dimensions of a 3-d array
integer(i4b) :: nc2flag(2), nc3flag(3), nc4flag(4)
!     nc2flag:   Vector with the 2 possible values of a flag variable
!     nc3flag:   Vector with the 3 possible values of a flag variable
!     nc4flag:   Vector with the 4 possible values of a flag variable
integer(i4b) :: nc1cor_i(1), nc1cor_j(1), &
                nc1cor_kc(1), nc1cor_kt(1), nc1cor_kr(1), &
                nc2cor_ij(2), &
                nc3cor_ijkc(3), nc3cor_ijkt(3), nc3cor_ijkr(3)
!     nc1cor(1): Corner of a 1-d array
!     nc2cor(2): Corner of a 2-d array
!     nc3cor(3): Corner of a 3-d array
integer(i4b) :: nc1cnt_i(1), nc1cnt_j(1), &
                nc1cnt_kc(1), nc1cnt_kt(1), nc1cnt_kr(1), &
                nc2cnt_ij(2), &
                nc3cnt_ijkc(3), nc3cnt_ijkt(3), nc3cnt_ijkr(3)
!     nc1cnt(1): Count of a 1-d array
!     nc2cnt(2): Count of a 2-d array
!     nc3cnt(3): Count of a 3-d array
character(len=256) :: buffer
character(len= 16), parameter :: filename_extension = '.nc'
character(len= 16), allocatable :: coord_id(:)

#else
errormsg = ' >>> output1: Parameter NETCDF must be either 1 or 2!'
call error(errormsg)
#endif

character(len=64), parameter :: thisroutine = 'output1'

#if (NETCDF==2)   /* time-slice file in NetCDF format */

nc1cor_i = (/ 1 /)
nc1cnt_i = (/ IMAX+1 /)

nc1cor_j = (/ 1 /)
nc1cnt_j = (/ JMAX+1 /)

nc1cor_kc = (/ 1 /)
nc1cnt_kc = (/ KCMAX+1 /)

nc1cor_kt = (/ 1 /)
nc1cnt_kt = (/ KTMAX+1 /)

nc1cor_kr = (/ 1 /)
nc1cnt_kr = (/ KRMAX+1 /)

nc2cor_ij = (/ 1, 1 /)
nc2cnt_ij = (/ IMAX+1, JMAX+1 /)

nc3cor_ijkc = (/ 1, 1, 1 /)
nc3cnt_ijkc = (/ IMAX+1, JMAX+1, KCMAX+1 /)

nc3cor_ijkt = (/ 1, 1, 1 /)
nc3cnt_ijkt = (/ IMAX+1, JMAX+1, KTMAX+1 /)

nc3cor_ijkr = (/ 1, 1, 1 /)
nc3cnt_ijkr = (/ IMAX+1, JMAX+1, KRMAX+1 /)

#endif

!-------- Create consecutively numbered file names --------

if (flag_3d_output) then
   ndat = ndat3d
else
   ndat = ndat2d
end if

if (ndat > 9999) then 
   errormsg = ' >>> output1: Too many time-slice files!'
   call error(errormsg)
endif

ndat_help  = ndat
ndat_1000s = ndat_help/1000
ndat_help  = ndat_help-ndat_1000s*1000
ndat_100s  = ndat_help/100
ndat_help  = ndat_help-ndat_100s*100
ndat_10s   = ndat_help/10
ndat_help  = ndat_help-ndat_10s*10
ndat_1s    = ndat_help

ch_1000s = char(ndat_1000s+ichar('0'))
ch_100s  = char(ndat_100s +ichar('0'))
ch_10s   = char(ndat_10s  +ichar('0'))
ch_1s    = char(ndat_1s   +ichar('0'))

if (flag_3d_output) then
   filename = trim(runname)//ch_1000s//ch_100s//ch_10s//ch_1s &
                           //trim(filename_extension)
else
   filename = trim(runname)//'_2d_'//ch_1000s//ch_100s//ch_10s//ch_1s &
                           //trim(filename_extension)
end if

filename_with_path = trim(OUTPATH)//'/'//trim(filename)

!-------- File initialization --------

#if (NETCDF==1)   /* time-slice file in native binary format */

!  ------ Open native binary file

open(unit=11, iostat=ios, file=trim(filename_with_path), status='new', &
              form='unformatted')

if (ios /= 0) then 
   errormsg = ' >>> output1: Error when opening an erg file!'
   call error(errormsg)
end if

!  ------ Global attributes

ch_attr_title = 'Time-slice output no. '//ch_1000s//ch_100s//ch_10s//ch_1s// &
                ' of simulation '//trim(runname)
write(unit=11) ch_attr_title

ch_attr_institution = 'Institute of Low Temperature Science, '// &
                      'Hokkaido University, Sapporo, Japan'
write(unit=11) ch_attr_institution

ch_attr_source = 'SICOPOLIS Version '//VERSION
write(unit=11) ch_attr_source

call date_and_time(ch_date, ch_time, ch_zone)
ch_attr_history = ch_date(1:4)//'-'//ch_date(5:6)//'-'//ch_date(7:8)//' '// &
                  ch_time(1:2)//':'//ch_time(3:4)//':'//ch_time(5:6)//' '// &
                  ch_zone(1:3)//':'//ch_zone(4:5)//' - Data produced'
write(unit=11) ch_attr_history

ch_attr_references = 'http://www.sicopolis.net/'
write(unit=11) ch_attr_references

#elif (NETCDF==2) /* time-slice file in NetCDF format */

if (allocated(coord_id)) deallocate(coord_id); allocate(coord_id(5))
coord_id(1) = 'x'; coord_id(2) = 'y'
coord_id(3) = 'zeta_c'; coord_id(4) = 'zeta_t'; coord_id(5) = 'zeta_r'

!  ------ Open NetCDF file

call check( nf90_create(trim(filename_with_path), NF90_NOCLOBBER, ncid), &
            thisroutine )

!  ------ Global attributes

buffer = 'Time-slice output no. '//ch_1000s//ch_100s//ch_10s//ch_1s// &
         ' of simulation '//trim(runname)
call check( nf90_put_att(ncid, NF90_GLOBAL, 'title', trim(buffer)), &
            thisroutine )

buffer = 'Institute of Low Temperature Science, Hokkaido University, '// &
         'Sapporo, Japan'
call check( nf90_put_att(ncid, NF90_GLOBAL, 'institution', trim(buffer)), &
            thisroutine )

buffer = 'SICOPOLIS Version '//VERSION
call check( nf90_put_att(ncid, NF90_GLOBAL, 'source', trim(buffer)), &
            thisroutine )

call date_and_time(ch_date, ch_time, ch_zone)
buffer = ch_date(1:4)//'-'//ch_date(5:6)//'-'//ch_date(7:8)//' '// &
         ch_time(1:2)//':'//ch_time(3:4)//':'//ch_time(5:6)//' '// &
         ch_zone(1:3)//':'//ch_zone(4:5)//' - Data produced'
call check( nf90_put_att(ncid, NF90_GLOBAL, 'history', trim(buffer)), &
            thisroutine )

buffer = 'http://www.sicopolis.net/'
call check( nf90_put_att(ncid, NF90_GLOBAL, 'references', trim(buffer)), &
            thisroutine )

!  ------ Definition of the dimensions

call check( nf90_def_dim(ncid, trim(coord_id(1)), IMAX+1,  ncd), thisroutine )
call check( nf90_def_dim(ncid, trim(coord_id(2)), JMAX+1,  ncd), thisroutine )
call check( nf90_def_dim(ncid, trim(coord_id(3)), KCMAX+1, ncd), thisroutine )
call check( nf90_def_dim(ncid, trim(coord_id(4)), KTMAX+1, ncd), thisroutine )
call check( nf90_def_dim(ncid, trim(coord_id(5)), KRMAX+1, ncd), thisroutine )

!  ------ Definition of the variables

!    ---- crs

call check( nf90_def_var(ncid, 'crs', NF90_BYTE, ncv), thisroutine ) 
#if (GRID==0 || GRID==1)
buffer = 'polar_stereographic'
call check( nf90_put_att(ncid, ncv, 'grid_mapping_name', trim(buffer)), &
            thisroutine )
#elif (GRID==2)
buffer = 'latitude_longitude'
call check( nf90_put_att(ncid, ncv, 'grid_mapping_name', trim(buffer)), &
            thisroutine )
#endif
#if (defined(ANT) \
     || defined(ASF) \
     || defined(GRL) \
     || defined(SCAND) \
     || defined(NHEM) \
     || defined(TIBET) \
     || defined(EMTP2SGE) \
     || defined(XYZ))   /* terrestrial ice sheet */
buffer = 'WGS84'
call check( nf90_put_att(ncid, ncv, 'ellipsoid', trim(buffer)), &
            thisroutine )
#elif (defined(NMARS) || defined(SMARS))   /* Martian ice sheet */
buffer = 'Mars_ellipsoid'
call check( nf90_put_att(ncid, ncv, 'ellipsoid', trim(buffer)), &
            thisroutine )
#else
errormsg = ' >>> output1: No valid domain (ANT, GRL etc.) specified!'
call error(errormsg)
#endif
#if (GRID==0 || GRID==1)
call check( nf90_put_att(ncid, ncv, 'false_easting', 0.0), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'false_northing', 0.0), &
            thisroutine )
lon0 = real(LAMBDA0*pi_180_inv,sp)
lon0 = modulo(lon0+180.0_sp, 360.0_sp)-180.0_sp
lon0 = real(nint(lon0*1.0e+04_sp),sp)*1.0e-04_sp
lat0 = real(PHI0*pi_180_inv,sp)
if (lat0 >  90.0_sp) lat0 =  90.0_sp
if (lat0 < -90.0_sp) lat0 = -90.0_sp
lat0 = real(nint(lat0*1.0e+04_sp),sp)*1.0e-04_sp
       ! reference longitude and standard parallel in deg rounded to 4 digits
if (lat0 >= 0.0_sp) then
   call check( nf90_put_att(ncid, ncv, &
                            'latitude_of_projection_origin',  90.0), &
                            thisroutine )
else
   call check( nf90_put_att(ncid, ncv, &
                            'latitude_of_projection_origin', -90.0), &
                            thisroutine )
end if
call check( nf90_put_att(ncid, ncv, &
                         'straight_vertical_longitude_from_pole', lon0), &
                         thisroutine )
call check( nf90_put_att(ncid, ncv, &
                         'standard_parallel', lat0), &
                         thisroutine )
#endif

!    ---- time

call check( nf90_def_var(ncid, 'time', NF90_FLOAT, ncv), &
            thisroutine )
buffer = 'a'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'time'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Time'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

if (forcing_flag == 1) then

!    ---- delta_ts

   call check( nf90_def_var(ncid, 'delta_ts', NF90_FLOAT, ncv), &
               thisroutine )
   buffer = 'degC'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'surface_temperature_anomaly'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Surface temperature anomaly'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

else if (forcing_flag == 2) then

!    ---- glac_index

   call check( nf90_def_var(ncid, 'glac_index', NF90_FLOAT, ncv), &
               thisroutine )
   buffer = '1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'glacial_index'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Glacial index'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

else if (forcing_flag == 3) then

!    ---- glac_index

   call check( nf90_def_var(ncid, 'glac_index', NF90_FLOAT, ncv), &
               thisroutine )
   buffer = '1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'glacial_index'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Glacial index'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   buffer = 'This variable will be assigned a dummy value only!'
   call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
               thisroutine )

end if

!    ---- z_sl

call check( nf90_def_var(ncid, 'z_sl', NF90_FLOAT, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'global_average_sea_level_change'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Sea level'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- V_tot

call check( nf90_def_var(ncid, 'V_tot', NF90_FLOAT, ncv), &
            thisroutine )
buffer = 'm3'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_volume'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ice volume'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- V_af

call check( nf90_def_var(ncid, 'V_af', NF90_FLOAT, ncv), &
            thisroutine )
buffer = 'm3'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_volume_not_displacing_sea_water'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ice volume above flotation'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- A_grounded

call check( nf90_def_var(ncid, 'A_grounded', NF90_FLOAT, ncv), &
            thisroutine )
buffer = 'm2'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'grounded_land_ice_area'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Area covered by grounded ice'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- A_floating

call check( nf90_def_var(ncid, 'A_floating', NF90_FLOAT, ncv), &
            thisroutine )
buffer = 'm2'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'floating_ice_shelf_area'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Area covered by floating ice'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- x (= xi)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc1d), &
            thisroutine )
call check( nf90_def_var(ncid, 'x', NF90_FLOAT, nc1d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'projection_x_coordinate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'x-coordinate of the grid point i'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'axis', 'x'), &
            thisroutine )

!    ---- y (= eta)

call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc1d), &
            thisroutine )
call check( nf90_def_var(ncid, 'y', NF90_FLOAT, nc1d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'projection_y_coordinate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'y-coordinate of the grid point j'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'axis', 'y'), &
            thisroutine )

!    ---- sigma_level_c

call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc1d), &
            thisroutine )
call check( nf90_def_var(ncid, 'sigma_level_c', NF90_FLOAT, nc1d, ncv), &
            thisroutine )
buffer = 'up'
call check( nf90_put_att(ncid, ncv, 'positive', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_kc_layer_sigma_coordinate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'sigma-coordinate of the grid point kc'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- sigma_level_t

call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc1d), &
            thisroutine )
call check( nf90_def_var(ncid, 'sigma_level_t', NF90_FLOAT, nc1d, ncv), &
            thisroutine )
buffer = 'up'
call check( nf90_put_att(ncid, ncv, 'positive', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_kt_layer_sigma_coordinate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'sigma-coordinate of the grid point kt'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- sigma_level_r

call check( nf90_inq_dimid(ncid, trim(coord_id(5)), nc1d), &
            thisroutine )
call check( nf90_def_var(ncid, 'sigma_level_r', NF90_FLOAT, nc1d, ncv), &
            thisroutine )
buffer = 'up'
call check( nf90_put_att(ncid, ncv, 'positive', trim(buffer)), &
            thisroutine )
buffer = 'lithosphere_layer_sigma_coordinate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'sigma-coordinate of the grid point kr'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- lon

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'lon', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'degrees_E'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'longitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Geographical longitude'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- lat

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'lat', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'degrees_N'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'latitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Geographical latitude'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- lambda

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'lambda', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'rad'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'longitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Geographical longitude'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- phi

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'phi', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'rad'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'latitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Geographical latitude'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- temp_s

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'temp_s', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'degC'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'surface_temperature'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Temperature at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- accum

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'prec', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_precipitation'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Annual precipitation at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- snowfall

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'snowfall', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_snowfall'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Annual snowfall at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- rainfall

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'rainfall', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_rainfall'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Annual rainfall at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- pdd

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'pdd', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'degC a'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_positive_degree_days'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Positive degree days at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- as_perp

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'as_perp', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_surface_mass_balance'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Mass balance at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- as_perp_apl

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'as_perp_apl', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'applied_land_ice_surface_mass_balance'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Applied mass balance at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- smb_corr

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'smb_corr', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_surface_mass_balance_diagnosed_correction'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Diagnosed correction of the mass balance at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

#if (DISC>0)   /* Ice discharge parameterisation */

!    ---- dis_perp

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dis_perp', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'ice_discharge'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ice discharge'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- cst_dist

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'cst_dist', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'km'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'coastal_distance'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Coastal distance'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- cos_grad_tc

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'cos_grad_tc', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = '1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'cos_alpha'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Cosine of angle between surface gradient and cst dist gradient'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- mask_mar

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'mask_mar', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = 'marginal_ring_mask'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Marginal ring mask'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
nc2flag = (/ 0, 1 /)
call check( nf90_put_att(ncid, ncv, 'flag_values', nc2flag), &
            thisroutine )
buffer = 'no_ring '// &
         'ring'
call check( nf90_put_att(ncid, ncv, 'flag_meanings', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

#endif

!    ---- q_geo

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'q_geo', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'W m-2'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'upward_geothermal_heat_flux_at_ground_level'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Geothermal heat flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- maske

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'maske', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = 'ice_land_sea_mask'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ice-land-sea mask'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
nc4flag = (/ 0, 1, 2, 3 /)
call check( nf90_put_att(ncid, ncv, 'flag_values', nc4flag), &
            thisroutine )
buffer = 'glaciated_land '// &
         'ice_free_land '// &
         'sea '// &
         'floating_ice'
call check( nf90_put_att(ncid, ncv, 'flag_meanings', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- maske_old

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'maske_old', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = 'ice_land_sea_mask_old'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ice-land-sea mask (old)'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
nc4flag = (/ 0, 1, 2, 3 /)
call check( nf90_put_att(ncid, ncv, 'flag_values', nc4flag), &
            thisroutine )
buffer = 'glaciated_land '// &
         'ice_free_land '// &
         'sea '// &
         'floating_ice'
call check( nf90_put_att(ncid, ncv, 'flag_meanings', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- mask_ablation_type

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'mask_ablation_type', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = 'mask_indicating_ablation_type'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Mask indicating ablation type'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
nc4flag = (/ -2, -1, 1, 2 /)
call check( nf90_put_att(ncid, ncv, 'flag_values', nc4flag), &
            thisroutine )
buffer = 'hidden (ocean) '// &
         'hidden (land) '// &
         'visible (land) '// &
         'visible (ocean)'
call check( nf90_put_att(ncid, ncv, 'flag_meanings', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- n_cts

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'n_cts', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = 'polythermal_condition_mask'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Mask for polythermal conditions'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
nc3flag = (/ -1, 0, 1 /)
call check( nf90_put_att(ncid, ncv, 'flag_values', nc3flag), &
            thisroutine )
buffer = 'cold_base '// &
         'temperate_base_with_cold_ice_above '// &
         'temperate_base_with_temperate_ice_above'
call check( nf90_put_att(ncid, ncv, 'flag_meanings', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- kc_cts

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'kc_cts', NF90_INT, nc2d, ncv), &
            thisroutine )
buffer = 'CTS_position_grid_index'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Grid index of the CTS position'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- zs

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'zs', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'surface_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Topography of the free surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- zm

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'zm', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'zm_interface_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Topography of the z=zm interface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- zb

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'zb', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'ice_base_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Topography of the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- zl

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'zl', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'bedrock_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Topography of the lithosphere surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- zl0

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'zl0', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'isostatically_relaxed_bedrock_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Topography of the isostatically relaxed lithosphere surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- H_cold

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'H_cold', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_cold_layer_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Thickness of the cold ice layer'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- H_temp

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'H_temp', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_temperate_layer_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Thickness of the temperate ice layer'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- H

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'H', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ice thickness'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- H_R

call check( nf90_def_var(ncid, 'H_R', NF90_FLOAT, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'lithosphere_layer_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Thickness of the lithosphere layer'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )

!    ---- Q_bm

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'Q_bm', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_basal_melt_rate'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Basal melting rate'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- Q_tld

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'Q_tld', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_temperate_layer_water_drainage'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Water drainage from the temperate layer'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- am_perp

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'am_perp', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_volume_flux_across_zm_interface'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Volume flux across the z=zm interface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- qx

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'qx', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm2 a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_vertical_integral_x_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal volume flux qx'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
buffer = 'Staggered grid variable, defined at (j,i+1/2)'
call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- qy

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'qy', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm2 a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_vertical_integral_y_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal volume flux qy'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
buffer = 'Staggered grid variable, defined at (j+1/2,i)'
call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- dzs_dt (= dzs_dtau)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dzs_dt', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'tendency_of_surface_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Rate of change of the topography of the free surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- dzm_dt (= dzm_dtau)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dzm_dt', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'tendency_of_zm_interface_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Rate of change of the topography of the z=zm interface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- dzb_dt (= dzb_dtau)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dzb_dt', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'tendency_of_ice_base_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Rate of change of the topography of the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- dzl_dt (= dzl_dtau)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dzl_dt', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'tendency_of_bedrock_altitude'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Rate of change of the topography of the lithosphere surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- dH_c_dt (= dH_c_dtau)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dH_c_dt', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'tendency_of_land_ice_kc_layer_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Rate of change of the thickness of the upper (kc) ice layer'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- dH_t_dt (= dH_t_dtau)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dH_t_dt', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'tendency_of_land_ice_kt_layer_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Rate of change of the thickness of the lower (kt) ice layer'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- dH_dt (= dH_dtau)

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'dH_dt', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'tendency_of_land_ice_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Rate of change of the ice thickness'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vx_b_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vx_b_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_base_x_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal velocity vx at the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vy_b_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vy_b_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_base_y_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal velocity vy at the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vz_b

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vz_b', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_base_z_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Vertical velocity vz at the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vh_b

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vh_b', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_base_horizontal_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal velocity vh at the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vx_s_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vx_s_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_surface_x_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal velocity vx at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vy_s_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vy_s_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_surface_y_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal velocity vy at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vz_s

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vz_s', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_surface_z_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Vertical velocity vz at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vh_s

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vh_s', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_surface_horizontal_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal velocity vh at the ice surface'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vx_m_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vx_m_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_vertical_mean_x_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Vertical mean of horizontal velocity vx'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vy_m_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vy_m_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_vertical_mean_y_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Vertical mean of horizontal velocity vy'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vh_m

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vh_m', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_vertical_mean_horizontal_velocity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Vertical mean of horizontal velocity vh'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- temp_b

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'temp_b', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'degC'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'basal_temperature'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Temperature at the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- temph_b

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'temph_b', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'degC'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'basal_temperature_rel_to_pmp'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Temperature at the ice base relative to the pressure melting point'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- tau_b_driving

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'tau_b_driving', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'Pa'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'magnitude_of_land_ice_driving_stress'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Driving stress'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- tau_b_drag

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'tau_b_drag', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'Pa'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'magnitude_of_land_ice_basal_drag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Basal drag'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- p_b_w

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'p_b_w', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'Pa'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'basal_water_pressure'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Basal water pressure'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- q_w

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'q_w', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm2 a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'basal_water_flux'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Volume flux of basal water'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- q_w_x

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'q_w_x', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm2 a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'basal_water_flux_x'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Volume flux of basal water in x-direction'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- q_w_y

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'q_w_y', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm2 a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'basal_water_flux_y'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Volume flux of basal water in y-direction'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- H_w

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'H_w', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'water_column_thickness'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Thickness of the water column under the ice base'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- q_gl_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'q_gl_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm2 a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_volume_flux_across_gl'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Horizontal volume flux across the grounding line'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- q_cf_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'q_cf_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'm a-1'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_volume_flux_due_to_calving'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Calving flux'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- ratio_sl_x

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'ratio_sl_x', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_x_slip_ratio'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ratio of basal to surface velocity (slip ratio) in x-direction, ' &
         // 'at (i+1/2,j)'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- ratio_sl_y

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'ratio_sl_y', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_y_slip_ratio'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Ratio of basal to surface velocity (slip ratio) in y-direction, ' &
         // 'at (i+1/2,j)'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_shelfy_stream_x

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_shelfy_stream_x', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
        thisroutine )
buffer = 'land_ice_x_shelfy_stream_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Shelfy stream flag in x-direction, at (i+1/2,j)'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
        thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_shelfy_stream_y

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_shelfy_stream_y', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_y_shelfy_stream_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Shelfy stream flag in y-direction, at (i,j+1/2)'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_shelfy_stream

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_shelfy_stream', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_shelfy_stream_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Shelfy stream flag'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_grounding_line_1

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_grounding_line_1', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_grounding_line_1_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Grounding line flag 1'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_grounding_line_2

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_grounding_line_2', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_grounding_line_2_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Grounding line flag 2'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_calving_front_1

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_calving_front_1', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_calving_front_1_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Calving front flag 1'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_calving_front_2

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_calving_front_2', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_calving_front_2_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Calving front flag 2'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_grounded_front_a_1

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_grounded_front_a_1', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_grounded_front_a_1_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Grounded front flag a 1'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_grounded_front_a_2

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_grounded_front_a_2', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_grounded_front_a_2_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Grounded front flag a 2'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_grounded_front_b_1

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_grounded_front_b_1', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_grounded_front_b_1_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Grounded front flag b 1'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- flag_grounded_front_b_2

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'flag_grounded_front_b_2', NF90_BYTE, nc2d, ncv), &
            thisroutine )
buffer = '-'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_grounded_front_b_2_flag'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Grounded front flag b 2'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vis_ave_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vis_ave_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'Pa s'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_vertical_average_viscosity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Depth-averaged viscosity (SIA/SSA)'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

!    ---- vis_int_g

call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc2d(1)), &
            thisroutine )
call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc2d(2)), &
            thisroutine )
call check( nf90_def_var(ncid, 'vis_int_g', NF90_FLOAT, nc2d, ncv), &
            thisroutine )
buffer = 'Pa s m'
call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
buffer = 'land_ice_vertical_integral_viscosity'
call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
            thisroutine )
buffer = 'Depth-integrated viscosity (SIA/SSA)'
call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
            thisroutine )
call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
            thisroutine )

if (flag_3d_output) then

!    ---- vx_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'vx_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'm a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_x_velocity'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Horizontal velocity vx in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   buffer = 'Staggered grid variable, defined at (kc,j,i+1/2)'
   call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- vy_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'vy_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'm a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_y_velocity'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Horizontal velocity vy in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   buffer = 'Staggered grid variable, defined at (kc,j+1/2,i)'
   call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- vz_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'vz_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'm a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_z_velocity'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Vertical velocity vz in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   buffer = 'Staggered grid variable, defined at (kc+1/2,j,i)'
   call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- vx_t

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'vx_t', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'm a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kt_layer_x_velocity'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Horizontal velocity vx in the lower (kt) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   buffer = 'Staggered grid variable, defined at (kt,j,i+1/2)'
   call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- vy_t

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'vy_t', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'm a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
            thisroutine )
   buffer = 'land_ice_kt_layer_y_velocity'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Horizontal velocity vy in the lower (kt) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   buffer = 'Staggered grid variable, defined at (kt,j+1/2,i)'
   call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- vz_t

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'vz_t', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'm a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kt_layer_z_velocity'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Vertical velocity vz in the lower (kt) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   buffer = 'Staggered grid variable, defined at (kt+1/2,j,i)'
   call check( nf90_put_att(ncid, ncv, 'comment', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- temp_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'temp_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'degC'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_temperature'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Temperature in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- omega_t

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'omega_t', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = '1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kt_layer_water_content'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Water content in the lower (kt) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- temp_r

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(5)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'temp_r', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'degC'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'lithosphere_layer_temperature'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Temperature in the lithosphere layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- enth_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'enth_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'J kg-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_enthalpy'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Enthalpy in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- enth_t

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'enth_t', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'J kg-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kt_layer_enthalpy'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Enthalpy in the lower (kt) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- omega_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'omega_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = '1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_water_content'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Water content in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- enh_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'enh_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = '1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_flow_enhancement_factor'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Flow enhancement factor in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- enh_t

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'enh_t', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = '1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kt_layer_flow_enhancement_factor'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Flow enhancement factor in the lower (kt) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- age_c

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(3)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'age_c', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'a'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kc_layer_age'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Age in the upper (kc) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )

!    ---- age_t

   call check( nf90_inq_dimid(ncid, trim(coord_id(1)), nc3d(1)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(2)), nc3d(2)), &
               thisroutine )
   call check( nf90_inq_dimid(ncid, trim(coord_id(4)), nc3d(3)), &
               thisroutine )
   call check( nf90_def_var(ncid, 'age_t', NF90_FLOAT, nc3d, ncv), &
               thisroutine )
   buffer = 'a'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_kt_layer_age'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Age in the lower (kt) ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'grid_mapping', 'crs'), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'coordinates', 'lat lon'), &
               thisroutine )
                               
end if

!    ---- End of the definitions

call check( nf90_enddef(ncid), thisroutine )

#endif

!-------- Ice thickness and time derivative --------

H       = H_c       + H_t
dH_dtau = dH_c_dtau + dH_t_dtau

!-------- Thickness of the cold and temperate layers --------

H_cold = 0.0_dp
H_temp = 0.0_dp

#if (CALCMOD==1)
do i=1, IMAX-1
do j=1, JMAX-1
   H_temp(j,i) = H_t(j,i)
end do
end do
#elif (CALCMOD==0 || CALCMOD==2 || CALCMOD==3 || CALCMOD==-1)
do i=1, IMAX-1
do j=1, JMAX-1
   H_temp(j,i) = H_c(j,i)*eaz_c_quotient(kc_cts(j,i))
end do
end do
#endif

H_cold = H - H_temp

!-------- Enthalpies for the non-enthalpy methods (POLY, COLD, ISOT) --------

#if (CALCMOD==1)

do i=0, IMAX
do j=0, JMAX

   do kc=0, KCMAX
      enth_c(kc,j,i) = enth_fct_temp_omega(temp_c(kc,j,i), 0.0_dp)
   end do

   if ( (maske(j,i)==0_i1b).and.(n_cts(j,i)==1_i1b) ) then
      do kt=0, KTMAX
         enth_t(kt,j,i) = enth_fct_temp_omega(temp_t_m(kt,j,i), omega_t(kt,j,i))
      end do
   else
      do kt=0, KTMAX
         enth_t(kt,j,i) = enth_c(0,j,i)
      end do
   end if

end do
end do

#elif (CALCMOD==0 || CALCMOD==-1)

do i=0, IMAX
do j=0, JMAX

   do kc=0, KCMAX
      enth_c(kc,j,i) = enth_fct_temp_omega(temp_c(kc,j,i), 0.0_dp)
   end do

   do kt=0, KTMAX
      enth_t(kt,j,i) = enth_c(0,j,i)
   end do

end do
end do

#endif

!-------- Vertical mean of horizontal velocities --------

vx_m_g = 0.0_dp
vy_m_g = 0.0_dp

do i=1, IMAX-1
do j=1, JMAX-1

   if ( (maske(j,i)==0_i1b).or.(maske(j,i)==3_i1b) ) then
      vx_m_g(j,i) = 0.5_dp*(vx_m(j,i)+vx_m(j,i-1))
      vy_m_g(j,i) = 0.5_dp*(vy_m(j,i)+vy_m(j-1,i))
   end if

end do
end do

!-------- Driving stress and basal drag --------

tau_b_driving = 0.0_dp
tau_b_drag    = 0.0_dp

do i=0, IMAX
do j=0, JMAX

   if (maske(j,i)==0_i1b) then   ! grounded ice

      tau_b_driving(j,i) = RHO*G*H(j,i) &
                           * sqrt( dzs_dxi_g(j,i)**2 + dzs_deta_g(j,i)**2 )

      if (.not.flag_shelfy_stream(j,i)) then
         tau_b_drag(j,i) = tau_b_driving(j,i)
      else
         tau_b_drag(j,i) = no_value_neg_2   ! dummy value
      end if

   else if (maske(j,i)==3_i1b) then   ! floating ice

      tau_b_driving(j,i) = RHO*G*H(j,i) &
                           * sqrt( dzs_dxi_g(j,i)**2 + dzs_deta_g(j,i)**2 )

      tau_b_drag(j,i)    = 0.0_dp

   end if

end do
end do

!-------- Computation of the ice volume, the volume above flotation,
!         the area covered by grounded ice, the area covered by floating ice,
!         the total surface mass balance, the total basal mass balance
!         and the total calving flux --------

rhosw_rho_ratio = RHO_SW/RHO

V_grounded = 0.0_dp
V_floating = 0.0_dp
V_gr_redu  = 0.0_dp
A_grounded = 0.0_dp
A_floating = 0.0_dp

do i=0, IMAX
do j=0, JMAX

   if (maske(j,i)==0_i1b) then   ! grounded ice

      V_grounded = V_grounded + H(j,i)*area(j,i)
      A_grounded = A_grounded + area(j,i)
      V_gr_redu  = V_gr_redu &
                   + rhosw_rho_ratio*max((z_sl-zl(j,i)),0.0_dp)*area(j,i)

   else if (maske(j,i)==3_i1b) then   ! floating ice

      V_floating = V_floating + H(j,i)*area(j,i)
      A_floating = A_floating + area(j,i)

   end if

end do
end do

V_tot   = V_grounded + V_floating
V_af    = V_grounded - V_gr_redu

!-------- Convert data to real*4 and seconds to years --------

#if (!defined(OUT_TIMES) || OUT_TIMES==1)
time_conv = real(time*sec_to_year,sp)
#elif (OUT_TIMES==2)
time_conv = real((time+year_zero)*sec_to_year,sp)
#else
errormsg = ' >>> output1: OUT_TIMES must be either 1 or 2!'
call error(errormsg)
#endif

delta_ts_conv   = real(delta_ts,sp)
glac_index_conv = real(glac_index,sp)
z_sl_conv       = real(z_sl,sp)
V_tot_conv      = real(V_tot,sp)
V_af_conv       = real(V_af,sp)
A_grounded_conv = real(A_grounded,sp)
A_floating_conv = real(A_floating,sp)
H_R_conv        = real(H_R,sp)

do i=0, IMAX
   xi_conv(i) = real(xi(i),sp)
end do

do j=0, JMAX
   eta_conv(j) = real(eta(j),sp)
end do

do kc=0, KCMAX
   sigma_level_c_conv(kc) = real(eaz_c_quotient(kc),sp)
end do

do kt=0, KTMAX
   sigma_level_t_conv(kt) = real(zeta_t(kt),sp)
end do

do kr=0, KRMAX
   sigma_level_r_conv(kr) = real(kr,sp)/real(KRMAX,sp)
end do

do i=0, IMAX
do j=0, JMAX

   maske_conv(i,j)              = maske(j,i)
   maske_old_conv(i,j)          = maske_old(j,i)
   mask_ablation_type_conv(i,j) = mask_ablation_type(j,i)
   n_cts_conv(i,j)              = n_cts(j,i)
   kc_cts_conv(i,j)             = kc_cts(j,i)

   lambda_conv(i,j)    = real(lambda(j,i),sp) 
   phi_conv(i,j)       = real(phi(j,i),sp) 
   lon_conv(i,j)       = real(lambda(j,i)*pi_180_inv,sp)   ! longitude in deg
   lon_conv(i,j)       = modulo(lon_conv(i,j)+180.0_sp, 360.0_sp)-180.0_sp
                                    ! mapping to interval [-180 deg, +180 deg)
   lat_conv(i,j)       = real(phi(j,i)   *pi_180_inv,sp)   ! latitute  in deg
   if (lat_conv(i,j) >  90.0_sp) lat_conv(i,j) =  90.0_sp
   if (lat_conv(i,j) < -90.0_sp) lat_conv(i,j) = -90.0_sp
                                 ! constraining to interval [-90 deg, +90 deg]
   temp_s_conv(i,j)      = real(temp_s(j,i),sp)
   accum_conv(i,j)       = real(accum(j,i)*YEAR_SEC,sp)
   snowfall_conv(i,j)    = real(snowfall(j,i)*YEAR_SEC,sp)
   rainfall_conv(i,j)    = real(rainfall(j,i)*YEAR_SEC,sp)
   pdd_conv(i,j)         = real(ET(j,i)*one_year,sp)
   as_perp_conv(i,j)     = real(as_perp(j,i)*YEAR_SEC,sp)
   as_perp_apl_conv(i,j) = real(as_perp_apl(j,i)*YEAR_SEC,sp)
   smb_corr_conv(i,j)    = real(smb_corr(j,i)*YEAR_SEC,sp)

#if (DISC>0)   /* Ice discharge parameterisation */
   dis_perp_conv(i,j)  = real(dis_perp(j,i)*YEAR_SEC,sp)
   cst_dist_conv(i,j)  = real(cst_dist(j,i)*0.001_dp,sp)
   cos_grad_tc_conv(i,j) = real(cos_grad_tc(j,i),sp)
   mask_mar_conv(i,j)  = mask_mar(j,i)
#endif

   q_geo_conv(i,j)     = real(q_geo(j,i),sp)
   zs_conv(i,j)        = real(zs(j,i),sp)
   zm_conv(i,j)        = real(zm(j,i),sp)
   zb_conv(i,j)        = real(zb(j,i),sp)
   zl_conv(i,j)        = real(zl(j,i),sp)
   zl0_conv(i,j)       = real(zl0(j,i),sp)
   H_cold_conv(i,j)    = real(H_cold(j,i),sp)
   H_temp_conv(i,j)    = real(H_temp(j,i),sp)
   H_conv(i,j)         = real(H(j,i),sp)
   Q_bm_conv(i,j)      = real(Q_bm(j,i)*YEAR_SEC,sp)
   Q_tld_conv(i,j)     = real(Q_tld(j,i)*YEAR_SEC,sp)
   am_perp_conv(i,j)   = real(am_perp(j,i)*YEAR_SEC,sp)
   qx_conv(i,j)        = real(qx(j,i)*YEAR_SEC,sp)
   qy_conv(i,j)        = real(qy(j,i)*YEAR_SEC,sp)
   dzs_dtau_conv(i,j)  = real(dzs_dtau(j,i)*YEAR_SEC,sp)
   dzm_dtau_conv(i,j)  = real(dzm_dtau(j,i)*YEAR_SEC,sp)
   dzb_dtau_conv(i,j)  = real(dzb_dtau(j,i)*YEAR_SEC,sp)
   dzl_dtau_conv(i,j)  = real(dzl_dtau(j,i)*YEAR_SEC,sp)
   dH_c_dtau_conv(i,j) = real(dH_c_dtau(j,i)*YEAR_SEC,sp)
   dH_t_dtau_conv(i,j) = real(dH_t_dtau(j,i)*YEAR_SEC,sp)
   dH_dtau_conv(i,j)   = real(dH_dtau(j,i)*YEAR_SEC,sp)
   vx_b_g_conv(i,j)    = real(vx_b_g(j,i)*YEAR_SEC,sp)
   vy_b_g_conv(i,j)    = real(vy_b_g(j,i)*YEAR_SEC,sp)
   vz_b_conv(i,j)      = real(vz_b(j,i)*YEAR_SEC,sp)
   vh_b_conv(i,j)      = sqrt( vx_b_g_conv(i,j)**2 + vy_b_g_conv(i,j)**2 )
   vx_s_g_conv(i,j)    = real(vx_s_g(j,i)*YEAR_SEC,sp)
   vy_s_g_conv(i,j)    = real(vy_s_g(j,i)*YEAR_SEC,sp)
   vz_s_conv(i,j)      = real(vz_s(j,i)*YEAR_SEC,sp)
   vh_s_conv(i,j)      = sqrt( vx_s_g_conv(i,j)**2 + vy_s_g_conv(i,j)**2 )
   vx_m_g_conv(i,j)    = real(vx_m_g(j,i)*YEAR_SEC,sp)
   vy_m_g_conv(i,j)    = real(vy_m_g(j,i)*YEAR_SEC,sp)
   vh_m_conv(i,j)      = sqrt( vx_m_g_conv(i,j)**2 + vy_m_g_conv(i,j)**2 )
   temp_b_conv(i,j)    = real(temp_b(j,i),sp)
   temph_b_conv(i,j)   = real(temph_b(j,i),sp)
   tau_b_driving_conv(i,j) = real(tau_b_driving(j,i),sp)
   tau_b_drag_conv(i,j)    = real(tau_b_drag(j,i),sp)
   p_b_w_conv(i,j)     = real(p_b_w(j,i),sp)
   q_w_conv(i,j)       = real(q_w(j,i)*YEAR_SEC,sp)
   q_w_x_conv(i,j)     = real(q_w_x(j,i)*YEAR_SEC,sp)
   q_w_y_conv(i,j)     = real(q_w_y(j,i)*YEAR_SEC,sp)
   H_w_conv(i,j)       = real(H_w(j,i),sp)
   q_gl_g_conv(i,j)    = real(q_gl_g(j,i)*YEAR_SEC,sp)
   q_cf_g_conv(i,j)    = real(calv_grounded(j,i)*YEAR_SEC,sp)
   ratio_sl_x_conv(i,j) = real(ratio_sl_x(j,i),sp)
   ratio_sl_y_conv(i,j) = real(ratio_sl_y(j,i),sp)

   if (flag_shelfy_stream_x(j,i)) then
      flag_shelfy_stream_x_conv(i,j) = 1_i1b
   else
      flag_shelfy_stream_x_conv(i,j) = 0_i1b
   end if

   if (flag_shelfy_stream_y(j,i)) then
      flag_shelfy_stream_y_conv(i,j) = 1_i1b
   else
      flag_shelfy_stream_y_conv(i,j) = 0_i1b
   end if

   if (flag_shelfy_stream(j,i)) then
      flag_shelfy_stream_conv(i,j) = 1_i1b
   else
      flag_shelfy_stream_conv(i,j) = 0_i1b
   end if

   if (flag_grounding_line_1(j,i)) then
      flag_grounding_line_1_conv(i,j) = 1_i1b
   else
      flag_grounding_line_1_conv(i,j) = 0_i1b
   end if

   if (flag_grounding_line_2(j,i)) then
      flag_grounding_line_2_conv(i,j) = 1_i1b
   else
      flag_grounding_line_2_conv(i,j) = 0_i1b
   end if

   if (flag_calving_front_1(j,i)) then
      flag_calving_front_1_conv(i,j) = 1_i1b
   else
      flag_calving_front_1_conv(i,j) = 0_i1b
   end if

   if (flag_calving_front_2(j,i)) then
      flag_calving_front_2_conv(i,j) = 1_i1b
   else
      flag_calving_front_2_conv(i,j) = 0_i1b
   end if

   if (flag_grounded_front_a_1(j,i)) then
      flag_grounded_front_a_1_conv(i,j) = 1_i1b
   else
      flag_grounded_front_a_1_conv(i,j) = 0_i1b
   end if

   if (flag_grounded_front_a_2(j,i)) then
      flag_grounded_front_a_2_conv(i,j) = 1_i1b
   else
      flag_grounded_front_a_2_conv(i,j) = 0_i1b
   end if

   if (flag_grounded_front_b_1(j,i)) then
      flag_grounded_front_b_1_conv(i,j) = 1_i1b
   else
      flag_grounded_front_b_1_conv(i,j) = 0_i1b
   end if

   if (flag_grounded_front_b_2(j,i)) then
      flag_grounded_front_b_2_conv(i,j) = 1_i1b
   else
      flag_grounded_front_b_2_conv(i,j) = 0_i1b
   end if

   vis_ave_g_conv(i,j) = real(vis_ave_g(j,i),sp)
   vis_int_g_conv(i,j) = real(vis_int_g(j,i),sp)

   do kr=0, KRMAX
      temp_r_conv(i,j,kr) = real(temp_r(kr,j,i),sp)
   end do

   do kt=0, KTMAX
      vx_t_conv(i,j,kt)    = real(vx_t(kt,j,i)*YEAR_SEC,sp)
      vy_t_conv(i,j,kt)    = real(vy_t(kt,j,i)*YEAR_SEC,sp)
      vz_t_conv(i,j,kt)    = real(vz_t(kt,j,i)*YEAR_SEC,sp)
      omega_t_conv(i,j,kt) = real(omega_t(kt,j,i),sp)
      age_t_conv(i,j,kt)   = real(age_t(kt,j,i)*sec_to_year,sp)
      enth_t_conv(i,j,kt)  = real(enth_t(kt,j,i),sp)
      enh_t_conv(i,j,kt)   = real(enh_t(kt,j,i),sp)
   end do

   do kc=0, KCMAX
      vx_c_conv(i,j,kc)    = real(vx_c(kc,j,i)*YEAR_SEC,sp)
      vy_c_conv(i,j,kc)    = real(vy_c(kc,j,i)*YEAR_SEC,sp)
      vz_c_conv(i,j,kc)    = real(vz_c(kc,j,i)*YEAR_SEC,sp)
      temp_c_conv(i,j,kc)  = real(temp_c(kc,j,i),sp)
      age_c_conv(i,j,kc)   = real(age_c(kc,j,i)*sec_to_year,sp)
      enth_c_conv(i,j,kc)  = real(enth_c(kc,j,i),sp)
      omega_c_conv(i,j,kc) = real(omega_c(kc,j,i),sp)
      enh_c_conv(i,j,kc)   = real(enh_c(kc,j,i),sp)
   end do

end do
end do

!-------- Write data on file --------

#if (NETCDF==1)   /* time-slice file in native binary format */

write(unit=11) time_conv
if (forcing_flag == 1) then
   write(unit=11) delta_ts_conv
else if (forcing_flag == 2) then
   write(unit=11) glac_index_conv
else if (forcing_flag == 3) then
   glac_index_conv = real(no_value_pos_1,sp)   ! dummy value
   write(unit=11) glac_index_conv
end if
write(unit=11) z_sl_conv

write(unit=11) V_tot_conv
write(unit=11) V_af_conv
write(unit=11) A_grounded_conv
write(unit=11) A_floating_conv

write(unit=11) xi_conv
write(unit=11) eta_conv
write(unit=11) sigma_level_c_conv
write(unit=11) sigma_level_t_conv
write(unit=11) sigma_level_r_conv
write(unit=11) lon_conv
write(unit=11) lat_conv
write(unit=11) lambda_conv
write(unit=11) phi_conv
write(unit=11) temp_s_conv
write(unit=11) accum_conv
write(unit=11) snowfall_conv
write(unit=11) rainfall_conv
write(unit=11) pdd_conv
write(unit=11) as_perp_conv
write(unit=11) as_perp_apl_conv
write(unit=11) smb_corr_conv

#if (DISC>0)   /* Ice discharge parameterisation */

write(unit=11) dis_perp_conv
write(unit=11) cst_dist_conv
write(unit=11) cos_grad_tc_conv
write(unit=11) mask_mar_conv

#endif

write(unit=11) q_geo_conv
write(unit=11) maske_conv
write(unit=11) maske_old_conv
write(unit=11) n_cts_conv
write(unit=11) kc_cts_conv
write(unit=11) zs_conv
write(unit=11) zm_conv
write(unit=11) zb_conv
write(unit=11) zl_conv
write(unit=11) zl0_conv
write(unit=11) H_cold_conv
write(unit=11) H_temp_conv
write(unit=11) H_conv
write(unit=11) H_R_conv
write(unit=11) Q_bm_conv
write(unit=11) Q_tld_conv
write(unit=11) am_perp_conv
write(unit=11) qx_conv
write(unit=11) qy_conv
write(unit=11) dzs_dtau_conv
write(unit=11) dzm_dtau_conv
write(unit=11) dzb_dtau_conv
write(unit=11) dzl_dtau_conv
write(unit=11) dH_c_dtau_conv
write(unit=11) dH_t_dtau_conv
write(unit=11) dH_dtau_conv
write(unit=11) vx_b_g_conv
write(unit=11) vy_b_g_conv
write(unit=11) vz_b_conv
write(unit=11) vh_b_conv
write(unit=11) vx_s_g_conv
write(unit=11) vy_s_g_conv
write(unit=11) vz_s_conv
write(unit=11) vh_s_conv
write(unit=11) vx_m_g_conv
write(unit=11) vy_m_g_conv
write(unit=11) vh_m_conv
write(unit=11) temp_b_conv
write(unit=11) temph_b_conv
write(unit=11) tau_b_driving_conv
write(unit=11) tau_b_drag_conv
write(unit=11) p_b_w_conv
write(unit=11) q_w_conv
write(unit=11) q_w_x_conv
write(unit=11) q_w_y_conv
write(unit=11) H_w_conv
write(unit=11) q_gl_g_conv
write(unit=11) q_cf_g_conv
write(unit=11) ratio_sl_x_conv
write(unit=11) ratio_sl_y_conv
write(unit=11) flag_shelfy_stream_x_conv
write(unit=11) flag_shelfy_stream_y_conv
write(unit=11) flag_shelfy_stream_conv
write(unit=11) flag_grounding_line_1_conv
write(unit=11) flag_grounding_line_2_conv
write(unit=11) flag_calving_front_1_conv
write(unit=11) flag_calving_front_2_conv
write(unit=11) flag_grounded_front_a_1_conv
write(unit=11) flag_grounded_front_a_2_conv
write(unit=11) flag_grounded_front_b_1_conv
write(unit=11) flag_grounded_front_b_2_conv
write(unit=11) vis_ave_g_conv
write(unit=11) vis_int_g_conv

if (flag_3d_output) then
   write(unit=11) vx_c_conv
   write(unit=11) vy_c_conv
   write(unit=11) vz_c_conv
   write(unit=11) vx_t_conv
   write(unit=11) vy_t_conv
   write(unit=11) vz_t_conv
   write(unit=11) temp_c_conv
   write(unit=11) omega_t_conv
   write(unit=11) temp_r_conv
   write(unit=11) enth_c_conv
   write(unit=11) enth_t_conv
   write(unit=11) omega_c_conv
   write(unit=11) enh_c_conv
   write(unit=11) enh_t_conv
   write(unit=11) age_c_conv
   write(unit=11) age_t_conv
end if

#elif (NETCDF==2) /* time-slice file in NetCDF format */

call check( nf90_inq_varid(ncid, 'crs', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, 0), thisroutine )

call check( nf90_inq_varid(ncid, 'time', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, time_conv), thisroutine )

if (forcing_flag == 1) then

   call check( nf90_inq_varid(ncid, 'delta_ts', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, delta_ts_conv), thisroutine )

else if (forcing_flag == 2) then

   call check( nf90_inq_varid(ncid, 'glac_index', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, glac_index_conv), thisroutine )

else if (forcing_flag == 3) then

   call check( nf90_inq_varid(ncid, 'glac_index', ncv), thisroutine )
   glac_index_conv = real(no_value_pos_1,sp)   ! dummy value
   call check( nf90_put_var(ncid, ncv, glac_index_conv), thisroutine )

end if

call check( nf90_inq_varid(ncid, 'z_sl', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, z_sl_conv), thisroutine )

call check( nf90_inq_varid(ncid, 'V_tot', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, V_tot_conv), thisroutine )

call check( nf90_inq_varid(ncid, 'V_af', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, V_af_conv), thisroutine )

call check( nf90_inq_varid(ncid, 'A_grounded', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, A_grounded_conv), thisroutine )

call check( nf90_inq_varid(ncid, 'A_floating', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, A_floating_conv), thisroutine )

call check( nf90_inq_varid(ncid, 'x', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, xi_conv, &
                         start=nc1cor_i, count=nc1cnt_i), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'y', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, eta_conv, &
                         start=nc1cor_j, count=nc1cnt_j), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'sigma_level_c', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, sigma_level_c_conv, &
                         start=nc1cor_kc, count=nc1cnt_kc), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'sigma_level_t', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, sigma_level_t_conv, &
                         start=nc1cor_kt, count=nc1cnt_kt), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'sigma_level_r', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, sigma_level_r_conv, &
                         start=nc1cor_kr, count=nc1cnt_kr), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'lon', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, lon_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'lat', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, lat_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'lambda', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, lambda_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'phi', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, phi_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'temp_s', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, temp_s_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'prec', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, accum_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'snowfall', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, snowfall_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'rainfall', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, rainfall_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'pdd', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, pdd_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'as_perp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, as_perp_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'as_perp_apl', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, as_perp_apl_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'smb_corr', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, smb_corr_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

#if (DISC>0)   /* Ice discharge parameterisation */

call check( nf90_inq_varid(ncid, 'dis_perp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dis_perp_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'cst_dist', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, cst_dist_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'cos_grad_tc', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, cos_grad_tc_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'mask_mar', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, mask_mar_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

#endif

call check( nf90_inq_varid(ncid, 'q_geo', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, q_geo_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'maske', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, maske_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'maske_old', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, maske_old_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'mask_ablation_type', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, mask_ablation_type_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'n_cts', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, n_cts_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'kc_cts', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, kc_cts_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'zs', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, zs_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'zm', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, zm_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'zb', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, zb_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'zl', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, zl_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'zl0', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, zl0_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'H_cold', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, H_cold_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'H_temp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, H_temp_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'H', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, H_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'H_R', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, H_R_conv), thisroutine )

call check( nf90_inq_varid(ncid, 'Q_bm', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, Q_bm_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'Q_tld', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, Q_tld_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'am_perp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, am_perp_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'qx', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, qx_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'qy', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, qy_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'dzs_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dzs_dtau_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'dzm_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dzm_dtau_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'dzb_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dzb_dtau_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'dzl_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dzl_dtau_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'dH_c_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dH_c_dtau_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'dH_t_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dH_t_dtau_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'dH_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, dH_dtau_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vx_b_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vx_b_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vy_b_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vy_b_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vz_b', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vz_b_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vh_b', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vh_b_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vx_s_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vx_s_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vy_s_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vy_s_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vz_s', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vz_s_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vh_s', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vh_s_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vx_m_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vx_m_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vy_m_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vy_m_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vh_m', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vh_m_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'temp_b', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, temp_b_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'temph_b', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, temph_b_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'tau_b_driving', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, tau_b_driving_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'tau_b_drag', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, tau_b_drag_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'p_b_w', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, p_b_w_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'q_w', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, q_w_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'q_w_x', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, q_w_x_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'q_w_y', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, q_w_y_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'H_w', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, H_w_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'q_gl_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, q_gl_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'q_cf_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, q_cf_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'ratio_sl_x', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, ratio_sl_x_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'ratio_sl_y', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, ratio_sl_y_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_shelfy_stream_x', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_shelfy_stream_x_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_shelfy_stream_y', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_shelfy_stream_y_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_shelfy_stream', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_shelfy_stream_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_grounding_line_1', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_grounding_line_1_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_grounding_line_2', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_grounding_line_2_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_calving_front_1', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_calving_front_1_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_calving_front_2', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_calving_front_2_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_grounded_front_a_1', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_grounded_front_a_1_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_grounded_front_a_2', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_grounded_front_a_2_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_grounded_front_b_1', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_grounded_front_b_1_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'flag_grounded_front_b_2', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, flag_grounded_front_b_2_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vis_ave_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vis_ave_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

call check( nf90_inq_varid(ncid, 'vis_int_g', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, vis_int_g_conv, &
                         start=nc2cor_ij, count=nc2cnt_ij), &
            thisroutine )

if (flag_3d_output) then

   call check( nf90_inq_varid(ncid, 'vx_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, vx_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'vy_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, vy_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'vz_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, vz_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'vx_t', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, vx_t_conv, &
                            start=nc3cor_ijkt, count=nc3cnt_ijkt), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'vy_t', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, vy_t_conv, &
                            start=nc3cor_ijkt, count=nc3cnt_ijkt), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'vz_t', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, vz_t_conv, &
                            start=nc3cor_ijkt, count=nc3cnt_ijkt), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'temp_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, temp_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'omega_t', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, omega_t_conv, &
                            start=nc3cor_ijkt, count=nc3cnt_ijkt), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'temp_r', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, temp_r_conv, &
                            start=nc3cor_ijkr, count=nc3cnt_ijkr), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'enth_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, enth_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'enth_t', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, enth_t_conv, &
                            start=nc3cor_ijkt, count=nc3cnt_ijkt), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'omega_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, omega_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'enh_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, enh_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'enh_t', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, enh_t_conv, &
                            start=nc3cor_ijkt, count=nc3cnt_ijkt), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'age_c', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, age_c_conv, &
                            start=nc3cor_ijkc, count=nc3cnt_ijkc), &
               thisroutine )

   call check( nf90_inq_varid(ncid, 'age_t', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, age_t_conv, &
                            start=nc3cor_ijkt, count=nc3cnt_ijkt), &
               thisroutine )

end if

#endif

!-------- Close file --------

#if (NETCDF==1)   /* time-slice file in native binary format */

close(unit=11, status='keep')

#elif (NETCDF==2) /* time-slice file in NetCDF format */

call check( nf90_sync(ncid),  thisroutine )
call check( nf90_close(ncid), thisroutine )

deallocate(coord_id)

#endif

!-------- Increase file counter --------

ndat = ndat+1

if (flag_3d_output) then
   ndat3d = ndat
else
   ndat2d = ndat
end if

end subroutine output1

!-------------------------------------------------------------------------------
!> Writing of time-series data on file in ASCII format
!! (and optionally in NetCDF format).
!<------------------------------------------------------------------------------
subroutine output2(time, dxi, deta, delta_ts, glac_index, z_sl)

#if (NETCDF>1)
  use netcdf
  use nc_check_m
#endif

#if (DISC>0)
  use discharge_workers_m, only: dT_glann, dT_sub
#endif

implicit none

real(dp), intent(in) :: time, dxi, deta, delta_ts, glac_index, z_sl

integer(i4b) :: i, j
integer(i4b) :: n_base, n_tempbase
real(dp) :: time_val, &
            V_tot, V_grounded, V_floating, A_tot, A_grounded, A_floating, &
            V_af, V_sle, V_temp, A_temp, &
            Q_s, Q_b, Q_temp, &
            H_max, H_t_max, zs_max, vs_max, Tbh_max, &
            accum_tot, calv_tot, mbp, &
            MB, LMT, OMT, LMB, OMB, LMH, OMH, ODT, ODH, &
            disc_lsc, disc_ssc
real(dp) :: x_pos, y_pos
real(dp), dimension(0:JMAX,0:IMAX) :: H, H_cold, H_temp
real(dp) :: V_gr_redu, A_surf, rhosw_rho_ratio
real(dp) :: vs_help, Tbh_help
real(dp) :: H_ave_sed, Tbh_ave_sed, Atb_sed
real(dp) :: sum_area_sed

#if (NETCDF>1)
integer(i4b), save :: ncid
integer(i4b)       :: ncd, ncv, nc1d
integer(i4b)       :: nc1cor(1), nc1cnt(1)
integer(i4b)       :: n_sync
real(dp)           :: dV_dt, precip_tot, runoff_tot, bmb_tot, mb_resid
real(dp), save     :: time_add_offset_val
character(len= 16) :: ch_date, ch_time, ch_zone
character(len=256) :: filename, filename_with_path, buffer
logical, save      :: grads_nc_tweaks
#endif

integer(i4b), save :: counter   = 0
logical,      save :: firstcall = .true.

character(len=64), parameter :: thisroutine = 'output2'

character(len=128), parameter :: &
                    fmt1 = '(1pe13.6,2(1pe13.4),/,' // &
                           '13x,6(1pe13.4),/,'      // &
                           '26x,3(1pe13.4),/,'      // &
                           '26x,5(1pe13.4),/)'

character(len=128), parameter :: &
                    fmt2 = '(1pe13.6,2(1pe13.4),/,' // &
                           '13x,3(1pe13.4),/)'

character(len=128), parameter :: &
                    fmt3 = '(1pe13.6,3(1pe13.4),/)'

counter = counter + 1

!-------- Ice thickness --------

H = H_c + H_t

!-------- Thickness of the cold and temperate layers --------

H_cold = 0.0_dp
H_temp = 0.0_dp

#if (CALCMOD==1)
do i=1, IMAX-1
do j=1, JMAX-1
   H_temp(j,i) = H_t(j,i)
end do
end do
#elif (CALCMOD==0 || CALCMOD==2 || CALCMOD==3 || CALCMOD==-1)
do i=1, IMAX-1
do j=1, JMAX-1
   H_temp(j,i) = H_c(j,i)*eaz_c_quotient(kc_cts(j,i))
end do
end do
#endif

H_cold = H - H_temp

!-------- Maximum ice elevation and thickness, ice volume,
!         sea-level equivalent, volume of the temperate ice,
!         area, area covered by temperate ice,
!         freshwater production due to melting and calving,
!         water drainage due to basal melting,
!         water drainage from the temperate layer,
!         maximum thickness of the temperate layer,
!         maximum surface velocity,
!         maximum basal temperature rel. to pmp --------

#if (defined(ANT) \
      || defined(GRL) \
      || defined(SCAND) \
      || defined(NHEM) \
      || defined(TIBET) \
      || defined(ASF) \
      || defined(EMTP2SGE) \
      || defined(XYZ))   /* terrestrial ice sheet */

rhosw_rho_ratio = RHO_SW/RHO

#elif (defined(NMARS) || defined(SMARS))   /* Martian ice sheet */

rhosw_rho_ratio = 0.0_dp   ! dummy value

#else
errormsg = ' >>> output2: No valid domain specified!'
call error(errormsg)
#endif

V_grounded = 0.0_dp
V_gr_redu  = 0.0_dp
V_floating = 0.0_dp
A_grounded = 0.0_dp
A_floating = 0.0_dp
V_temp     = 0.0_dp
A_temp     = 0.0_dp
Q_s        = 0.0_dp
Q_b        = 0.0_dp
Q_temp     = 0.0_dp
H_max      = 0.0_dp
H_t_max    = 0.0_dp
zs_max     = no_value_neg_2
vs_max     = 0.0_dp
Tbh_max    = no_value_neg_2
accum_tot  = 0.0_dp

do i=1, IMAX-1
do j=1, JMAX-1

   if (maske(j,i)==0_i1b) then   ! grounded ice

      if (zs(j,i)       > zs_max)  zs_max  = zs(j,i)
      if (H(j,i)        > H_max  ) H_max   = H(j,i)
      if (H_temp(j,i)   > H_t_max) H_t_max = H_temp(j,i)

      V_grounded = V_grounded + H(j,i)     *area(j,i)
      V_temp     = V_temp     + H_temp(j,i)*area(j,i)

#if (defined(ANT) \
      || defined(GRL) \
      || defined(SCAND) \
      || defined(NHEM) \
      || defined(TIBET) \
      || defined(ASF) \
      || defined(EMTP2SGE) \
      || defined(XYZ))   /* terrestrial ice sheet */

      V_gr_redu = V_gr_redu &
                  + rhosw_rho_ratio*max((z_sl-zl(j,i)),0.0_dp)*area(j,i)

#endif

      A_grounded = A_grounded + area(j,i)

      if (n_cts(j,i) /= -1_i1b) A_temp = A_temp + area(j,i)

      vs_help = sqrt(0.25_dp &
                       * ( (vx_c(KCMAX,j,i)+vx_c(KCMAX,j,i-1))**2 &
                          +(vy_c(KCMAX,j,i)+vy_c(KCMAX,j-1,i))**2 ) )
      if (vs_help > vs_max) vs_max = vs_help

      if (n_cts(j,i) >= 0_i1b) then   ! temperate base
         Tbh_max = 0.0_dp
      else   ! cold base
         Tbh_help = min((temp_c(0,j,i)-temp_c_m(0,j,i)), 0.0_dp)
         if (Tbh_help > Tbh_max) Tbh_max = Tbh_help
      end if

      accum_tot = accum_tot + accum(j,i)*area(j,i)

   else if (maske(j,i)==3_i1b) then   ! floating ice
                                      ! (basal temperature assumed to be below
                                      ! the pressure melting point for pure ice)

      if (zs(j,i)    > zs_max) zs_max = zs(j,i)
      if (H(j,i)     > H_max)  H_max  = H(j,i)

      V_floating = V_floating + H(j,i)*area(j,i)
      A_floating = A_floating + area(j,i)

      vs_help = sqrt(0.25_dp &
                       * ( (vx_c(KCMAX,j,i)+vx_c(KCMAX,j,i-1))**2 &
                          +(vy_c(KCMAX,j,i)+vy_c(KCMAX,j-1,i))**2 ) )
      if (vs_help > vs_max) vs_max = vs_help

      Tbh_help = min((temp_c(0,j,i)-temp_c_m(0,j,i)), 0.0_dp)
      if (Tbh_help > Tbh_max) Tbh_max = Tbh_help

      accum_tot = accum_tot + accum(j,i)*area(j,i)

   end if

   if (     (maske(j,i)==0_i1b).or.(maske_old(j,i)==0_i1b) &
        .or.(maske(j,i)==3_i1b).or.(maske_old(j,i)==3_i1b) ) then
                     ! grounded or floating ice before or after the time step

      Q_s   = Q_s  + as_perp_apl(j,i) * area(j,i)
      Q_b   = Q_b  + Q_bm(j,i)        * area(j,i)

   end if

   if ( (maske(j,i)==0_i1b).or.(maske_old(j,i)==0_i1b) ) then
                     ! grounded ice before or after the time step

      Q_temp = Q_temp + Q_tld(j,i) * area(j,i)

   end if

end do
end do

!-------- Conversion --------

#if (defined(ANT) \
      || defined(GRL) \
      || defined(SCAND) \
      || defined(NHEM) \
      || defined(TIBET) \
      || defined(ASF) \
      || defined(EMTP2SGE) \
      || defined(XYZ))   /* terrestrial ice sheet */

A_surf = 3.61132e+14_dp              ! global ocean area, in m^2

V_af   = V_grounded - V_gr_redu
V_sle  = V_af*(RHO/RHO_W)/A_surf     ! m^3 ice equiv./m^2 -> m water equiv.

#elif (defined(NMARS) || defined(SMARS))   /* Martian ice sheet */

A_surf = 1.4441e+14_dp            ! surface area of Mars, in m^2
V_sle  = V_grounded*(RHO_I/RHO_W)*(1.0_dp-FRAC_DUST)/A_surf
                                  ! m^3 (ice+dust) equiv./m^2 -> m water equiv.

#endif

#if (!defined(OUT_TIMES) || OUT_TIMES==1)
time_val = time *sec_to_year   ! s -> a
#elif (OUT_TIMES==2)
time_val = (time+year_zero) *sec_to_year   ! s -> a
#else
errormsg = ' >>> output2: OUT_TIMES must be either 1 or 2!'
call error(errormsg)
#endif

V_tot = V_grounded + V_floating   ! in m^3
A_tot = A_grounded + A_floating   ! in m^2

vs_max = vs_max *YEAR_SEC              ! m/s -> m/a

#if (defined(ANT) \
      || defined(GRL) \
      || defined(SCAND) \
      || defined(NHEM) \
      || defined(TIBET) \
      || defined(ASF) \
      || defined(EMTP2SGE) \
      || defined(XYZ))   /* terrestrial ice sheet */

Q_s    = Q_s    *YEAR_SEC *(RHO/RHO_W)
                ! m^3/s ice equiv. -> m^3/a water equiv.
Q_b    = Q_b    *YEAR_SEC *(RHO/RHO_W)
                ! m^3/s ice equiv. -> m^3/a water equiv.
Q_temp = Q_temp *YEAR_SEC *(RHO/RHO_W)
                ! m^3/s ice equiv. -> m^3/a water equiv.

accum_tot = accum_tot    *YEAR_SEC *(RHO/RHO_W)
                ! m^3/s ice equiv. -> m^3/a water equiv.

#elif (defined(NMARS) || defined(SMARS))   /* Martian ice sheet */

Q_s    = Q_s    *YEAR_SEC *(RHO_I/RHO_W)*(1.0_dp-FRAC_DUST)
                ! m^3/s (ice+dust) equiv. -> m^3/a water equiv.
Q_b    = Q_b    *YEAR_SEC *(RHO_I/RHO_W)*(1.0_dp-FRAC_DUST)
                ! m^3/s (ice+dust) equiv. -> m^3/a water equiv.
Q_temp = Q_temp *YEAR_SEC *(RHO_I/RHO_W)*(1.0_dp-FRAC_DUST)
                ! m^3/s (ice+dust) equiv. -> m^3/a water equiv.
#endif

!-------- Writing of data on time-series file --------

if (forcing_flag == 1) then

   write(unit=12, fmt=trim(fmt1)) time_val, delta_ts, z_sl, &
      V_tot, V_grounded, V_floating, A_tot, A_grounded, A_floating, &
      V_sle, V_temp, A_temp, &
      H_max, H_t_max, zs_max, vs_max, Tbh_max

else if (forcing_flag == 2) then

   write(unit=12, fmt=trim(fmt1)) time_val, glac_index, z_sl, &
      V_tot, V_grounded, V_floating, A_tot, A_grounded, A_floating, &
      V_sle, V_temp, A_temp, &
      H_max, H_t_max, zs_max, vs_max, Tbh_max

else if (forcing_flag == 3) then

   write(unit=12, fmt=trim(fmt1)) time_val, no_value_pos_1, z_sl, &
      V_tot, V_grounded, V_floating, A_tot, A_grounded, A_floating, &
      V_sle, V_temp, A_temp, &
      H_max, H_t_max, zs_max, vs_max, Tbh_max

end if

!-------- Special output --------

!  ------ Time-series data for the sediment region of HEINO

#if (defined(XYZ))

#if (defined(HEINO))

!    ---- Average ice thickness, average basal temperature rel. to pmp,
!         temperate basal area

H_ave_sed    = 0.0_dp
Tbh_ave_sed  = 0.0_dp
Atb_sed      = 0.0_dp
sum_area_sed = 0.0_dp

do i=1, IMAX-1
do j=1, JMAX-1

   if (maske_sedi(j,i)==7_i1b) then   ! sediment region

      sum_area_sed = sum_area_sed + area(j,i)

      H_ave_sed = H_ave_sed + area(j,i)*H(j,i)

      if (n_cts(j,i) /= -1_i1b) then   ! temperate base
         Tbh_help    = 0.0_dp
      else   ! cold base
         Tbh_help    = min((temp_c(0,j,i)-temp_c_m(0,j,i)), 0.0_dp)
      end if
      Tbh_ave_sed = Tbh_ave_sed + area(j,i)*Tbh_help

      if (n_cts(j,i) /= -1_i1b) Atb_sed = Atb_sed + area(j,i)

   end if

end do
end do

if (sum_area_sed > eps) then
   H_ave_sed   = H_ave_sed   / sum_area_sed
   Tbh_ave_sed = Tbh_ave_sed / sum_area_sed
else
   errormsg = ' >>> output2: No sediment area found!'
   call error(errormsg)
end if

!    ---- Writing of data on time-series file

if (forcing_flag == 1) then
   write(unit=15, fmt=trim(fmt2)) time_val, delta_ts, z_sl, &
                                  H_ave_sed, Tbh_ave_sed, Atb_sed
else if (forcing_flag == 2) then
   write(unit=15, fmt=trim(fmt2)) time_val, glac_index, z_sl, &
                                  H_ave_sed, Tbh_ave_sed, Atb_sed
else if (forcing_flag == 3) then
   write(unit=15, fmt=trim(fmt2)) time_val, no_value_pos_1, z_sl, &
                                  H_ave_sed, Tbh_ave_sed, Atb_sed
end if

#endif

#endif

!-------- Extended time-series file in NetCDF format --------

#if (NETCDF>1)

if (firstcall) then

!  ------ Open NetCDF file

   filename           = trim(RUNNAME)//'_ser.nc'
   filename_with_path = trim(OUTPATH)//'/'//trim(filename)

   call check( nf90_create(trim(filename_with_path), NF90_NOCLOBBER, ncid), &
               thisroutine )

   ncid_ser = ncid

!  ------ Global attributes

   buffer = 'Time-series output of simulation '//trim(RUNNAME)
   call check( nf90_put_att(ncid, NF90_GLOBAL, 'title', trim(buffer)), &
               thisroutine )

   buffer = 'Institute of Low Temperature Science, Hokkaido University, '// &
            'Sapporo, Japan'
   call check( nf90_put_att(ncid, NF90_GLOBAL, 'institution', trim(buffer)), &
               thisroutine )

   buffer = 'SICOPOLIS Version '//VERSION
   call check( nf90_put_att(ncid, NF90_GLOBAL, 'source', trim(buffer)), &
               thisroutine )

   call date_and_time(ch_date, ch_time, ch_zone)
   buffer = ch_date(1:4)//'-'//ch_date(5:6)//'-'//ch_date(7:8)//' '// &
            ch_time(1:2)//':'//ch_time(3:4)//':'//ch_time(5:6)//' '// &
            ch_zone(1:3)//':'//ch_zone(4:5)//' - Data produced'
   call check( nf90_put_att(ncid, NF90_GLOBAL, 'history', trim(buffer)), &
               thisroutine )

   buffer = 'http://www.sicopolis.net/'
   call check( nf90_put_att(ncid, NF90_GLOBAL, 'references', trim(buffer)), &
               thisroutine )

!  ------ Definition of the dimensions

   call set_grads_nc_tweaks(grads_nc_tweaks)

   if (grads_nc_tweaks) then
      call check( nf90_def_dim(ncid, 'x', 1, ncd), thisroutine )
      call check( nf90_def_dim(ncid, 'y', 1, ncd), thisroutine )
   end if

   call check( nf90_def_dim(ncid, 't', NF90_UNLIMITED, ncd), thisroutine )

!  ------ Definition of the variables

   if (grads_nc_tweaks) then

!    ---- x

      call check( nf90_inq_dimid(ncid, 'x', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 'x', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = 'm'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'x'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Dummy x-coordinate for one point'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )
      call check( nf90_put_att(ncid, ncv, 'axis', 'x'), thisroutine )

!    ---- y

      call check( nf90_inq_dimid(ncid, 'y', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 'y', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = 'm'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'y'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Dummy y-coordinate for one point'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )
      call check( nf90_put_att(ncid, ncv, 'axis', 'y'), thisroutine )

   end if

!    ---- Time

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 't', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'a'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'time'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Time'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )
   call check( nf90_put_att(ncid, ncv, 'axis', 't'), thisroutine )

   if (grads_nc_tweaks) then

!    ---- Time offset

      call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 't_add_offset', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = 'a'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'time_add_offset'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Time offset'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

      time_add_offset_val = min(time_val, 0.0_dp)

   end if

   if (forcing_flag == 1) then

!    ---- delta_ts

      call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 'delta_ts', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = 'degC'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'surface_temperature_anomaly'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Surface temperature anomaly'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

   else if ((forcing_flag == 2).or.(forcing_flag == 3)) then

!    ---- glac_index

      call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 'glac_index', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = '1'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'glacial_index'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Glacial index'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

   end if

!    ---- z_sl

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'z_sl', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'global_average_sea_level_change'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Sea level'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- V_tot

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'V_tot', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_volume'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Ice volume'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- V_grounded

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'V_grounded', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'grounded_land_ice_volume'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Volume of grounded ice'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- V_floating

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'V_floating', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'floating_ice_shelf_volume'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Volume of floating ice'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- A_tot

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'A_tot', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm2'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_area'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Ice area'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- A_grounded

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'A_grounded', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm2'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'grounded_land_ice_area'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Area covered by grounded ice'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- A_floating

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'A_floating', NF90_FLOAT, nc1d, ncv), &
            thisroutine )
   buffer = 'm2'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'floating_ice_shelf_area'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Area covered by floating ice'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- V_af

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'V_af', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_volume_not_displacing_sea_water'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Ice volume above flotation'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- V_sle

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'V_sle', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm SLE'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'land_ice_volume_sle'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Ice volume in SLE'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- V_temp

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'V_temp', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'temperate_land_ice_volume'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Volume of temperate ice'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- A_temp

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'A_temp', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm2'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'temperate_land_ice_area'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Area covered by temperate ice'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- Q_s

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'Q_s', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'tendency_of_land_ice_volume_due_to_surface_mass_balance'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Total surface mass balance'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- precip_tot

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'precip_tot', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'precipitation_rate'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Precipitation rate'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- runoff_tot

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'runoff_tot', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'runoff_rate'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Runoff rate'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- bmb_tot

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'bmb_tot', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'tendency_of_land_ice_volume_due_to_basal_mass_balance'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Total basal mass balance'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- Q_b

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'Q_b', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'basal_melting_rate'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Basal melting rate'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- Q_temp

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'Q_temp', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'temperate_layer_drainage_rate'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Drainage rate from the temperate ice layer'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- calv_tot

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'calv_tot', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'tendency_of_land_ice_volume_due_to_calving'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Total calving rate'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

#if (DISC>0)

!    ---- disc_lsc

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'disc_lsc', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'large_scale_ice_lost_into_the_ocean'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Large scale ice lost into the ocean'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- disc_ssc

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'disc_ssc', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'small_scale_ice_lost_into_the_ocean'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Small scale ice lost into the ocean'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- dT_glann

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'dT_glann', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'degC'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'global_annual_temperature_anomaly'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Global annual temperature anomaly'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- dT_sub

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'dT_sub', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'degC'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'subsurface_ocean_temperature_anomaly'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Subsurface ocean temperature anomaly'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

#endif

!    ---- dV_dt

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'dV_dt', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'tendency_of_land_ice_volume'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Rate of ice volume change'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- mb_resid

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'mb_resid', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm3 ice equiv. a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'total_mass_balance_residual'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Residual of the total mass balance'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- mbp

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'mbp', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = '1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'mass_balance_partition'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Mass balance partition'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- H_max

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'H_max', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'maximum_ice_thickness'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Maximum ice thickness'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- H_t_max

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'H_t_max', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'maximum_thickness_of_temperate_ice'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Maximum thickness of temperate ice'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- zs_max

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'zs_max', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'maximum_surface_elevation'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Maximum surface elevation'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- vs_max

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'vs_max', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'm a-1'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'maximum_surface_speed'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Maximum surface speed'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- Tbh_max

   call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
   call check( nf90_def_var(ncid, 'Tbh_max', NF90_FLOAT, nc1d, ncv), &
               thisroutine )
   buffer = 'degC'
   call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
               thisroutine )
   buffer = 'maximum_basal_temperature_relative_to_pmp'
   call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
               thisroutine )
   buffer = 'Maximum basal temperature relative to pmp'
   call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
               thisroutine )

!    ---- End of the definitions

   call check( nf90_enddef(ncid), thisroutine )

end if

!  ------ Computation of further data

! Q_s, Q_b, Q_temp from earlier computations (now in ice equiv.) 
Q_s    = Q_s    *(RHO_W/RHO)
         ! m^3/a water equiv. -> m^3/a ice equiv.
Q_b    = Q_b    *(RHO_W/RHO)
         ! m^3/a water equiv. -> m^3/a ice equiv.
Q_temp = Q_temp *(RHO_W/RHO)
         ! m^3/a water equiv. -> m^3/a ice equiv.

! Computation of volume fluxes

dV_dt      = 0.0_dp
precip_tot = 0.0_dp

do i=1, IMAX-1
do j=1, JMAX-1

   if (     (maske(j,i)==0_i1b).or.(maske_old(j,i)==0_i1b) &
        .or.(maske(j,i)==3_i1b).or.(maske_old(j,i)==3_i1b) ) then
                     ! grounded or floating ice before or after the time step

     ! Volume change
      dV_dt = dV_dt + (dzs_dtau(j,i)-dzb_dtau(j,i))*area(j,i)
      precip_tot    = precip_tot + accum(j,i)*area(j,i)

   end if

end do
end do

precip_tot = precip_tot *YEAR_SEC
             ! m^3/s ice equiv. -> m^3/a ice equiv.
dV_dt      = dV_dt      *YEAR_SEC
             ! m^3/s ice equiv. -> m^3/a ice equiv.

#if (MB_ACCOUNT==0)

runoff_tot = 0.0_dp
calv_tot   = 0.0_dp

do i=1, IMAX-1
do j=1, JMAX-1

   if (     (maske(j,i)==0_i1b).or.(maske_old(j,i)==0_i1b) &
        .or.(maske(j,i)==3_i1b).or.(maske_old(j,i)==3_i1b) ) then
                     ! grounded or floating ice before or after the time step

      runoff_tot = runoff_tot + runoff(j,i)*area(j,i)
      calv_tot   = calv_tot   + calv_grounded(j,i)*area(j,i)

   end if

end do
end do

bmb_tot = -(Q_b+Q_temp)   ! positive for supply, negative for loss

#elif (MB_ACCOUNT==1)

! MB:  total mass balance as computed in subroutine apply_smb
! LMT: total lost on land at the top of the ice sheet
! LMB: total lost on land at the base of the ice sheet
! LMH: total hidden lost on land
! OMT: total lost in the ocean  at the top of the ice sheet
! OMB: total lost in the ocean at the base of the ice sheet
! OMH: total hidden lost in the ocean
! ODT: total lost through ice discharge parameterization
! ODH: total hidden lost through ice discharge parameterization 

MB  = 0.0_dp
LMT = 0.0_dp;   OMT = 0.0_dp
LMB = 0.0_dp;   OMB = 0.0_dp
LMH = 0.0_dp;   OMH = 0.0_dp
ODT = 0.0_dp;   ODH = 0.0_dp

do i=1, IMAX-1
do j=1, JMAX-1

   if ( mask_ablation_type(j,i) /= 0_i1b ) then
                     ! glaciated land and ocean (including hidden melt points)

   ! Quantify what types of melt occurred
     select case ( mask_ablation_type(j,i) )
       case( 2_i1b )
         OMT = OMT + runoff_apl(j,i) * area(j,i)
         ODT = ODT + calv_grounded_apl(j,i) * area(j,i)
         OMB = OMB + Q_b_apl(j,i) * area(j,i)
       case( 1_i1b )
         LMT = LMT + runoff_apl(j,i) * area(j,i)
         ODT = ODT + calv_grounded_apl(j,i) * area(j,i)    ! ok
         LMB = LMB + Q_b_apl(j,i) * area(j,i)
       case( -1_i1b )
         ODH = ODH + calv_grounded_apl(j,i) * area(j,i)
         LMH = LMH + runoff_apl(j,i) * area(j,i)
!        note: Q_b_apl(j,i) not counted as hidden yet
       case( -2_i1b )
         ODH = ODH + calv_grounded_apl(j,i) * area(j,i)
         OMH = OMH + runoff_apl(j,i) * area(j,i)
!        note: Q_b_apl(j,i) not counted as hidden yet
     end select
        
   end if
      
   ! Actual ice mass balance (from top melt, bottom melt and calving)
   MB = MB + mb_source_apl(j,i)*area(j,i)

end do
end do

! Runoff on land (excluding basal melt)
runoff_tot     = LMT + LMH
! Ice discharge (excluding basal melt)
calv_tot       = OMT + OMH + ODT + ODH
! Ice discharge from ice flow, large scale
disc_lsc       = OMT + OMH
! Ice discharge from parameterization, small scale
disc_ssc       = ODT + ODH
! Basal mass balance (here basal melt only)
bmb_tot        = -LMB-OMB

MB            = MB     * YEAR_SEC
                ! m^3/s ice equiv. -> m^3/a ice equiv.
disc_lsc    = disc_lsc * YEAR_SEC
                ! m^3/s ice equiv. -> m^3/a ice equiv.
disc_ssc    = disc_ssc * YEAR_SEC
                ! m^3/s ice equiv. -> m^3/a ice equiv.
bmb_tot     = bmb_tot  * YEAR_SEC
                ! m^3/s ice equiv. -> m^3/a ice equiv.

#endif   /* if (MB_ACCOUNT==0) elif (MB_ACCOUNT==1) */

runoff_tot = runoff_tot *YEAR_SEC
                ! m^3/s ice equiv. -> m^3/a ice equiv.
calv_tot   = calv_tot   * YEAR_SEC
                ! m^3/s ice equiv. -> m^3/a ice equiv.

if (precip_tot /= 0.0_dp) then
   mbp = calv_tot/precip_tot
else
   mbp = 0.0_dp
end if

#if (MB_ACCOUNT==0)
mb_resid = Q_s + bmb_tot - calv_tot - dV_dt
#elif(MB_ACCOUNT==1)
mb_resid = MB - dV_dt
#endif

!  ------ Writing of data on NetCDF file

if (firstcall.and.grads_nc_tweaks) then

   nc1cor = (/ 1 /)

   call check( nf90_inq_varid(ncid, 'x', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, 0.0_sp, start=nc1cor), thisroutine )

   call check( nf90_inq_varid(ncid, 'y', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, 0.0_sp, start=nc1cor), thisroutine )

   call check( nf90_sync(ncid), thisroutine )

end if

nc1cor(1) = counter
! nc1cnt(1) = 1   ! (not needed, and causes troubles for whatever reason)

call check( nf90_inq_varid(ncid, 't', ncv), thisroutine )

if (.not.grads_nc_tweaks) then
   call check( nf90_put_var(ncid, ncv, real(time_val,sp), &
                            start=nc1cor), thisroutine )
else
   call check( nf90_put_var(ncid, ncv, &
                            real(time_val-time_add_offset_val,sp), &
                            start=nc1cor), thisroutine )
                         ! this makes sure that all times are >=0
                         ! (GrADS doesn't like negative numbers)
   call check( nf90_inq_varid(ncid, 't_add_offset', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, &
                            real(time_add_offset_val,sp), &
                            start=nc1cor), thisroutine )
end if

if (forcing_flag == 1) then

   call check( nf90_inq_varid(ncid, 'delta_ts', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(delta_ts,sp), &
                            start=nc1cor), thisroutine )

else if (forcing_flag == 2) then

   call check( nf90_inq_varid(ncid, 'glac_index', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(glac_index,sp), &
                            start=nc1cor), thisroutine )

else if (forcing_flag == 3) then

   call check( nf90_inq_varid(ncid, 'glac_index', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(no_value_neg_2,sp), &
                            start=nc1cor), thisroutine )

end if

call check( nf90_inq_varid(ncid, 'z_sl', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(z_sl,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'V_tot', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(V_tot,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'V_grounded', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(V_grounded,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'V_floating', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(V_floating,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'A_tot', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(A_tot,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'A_grounded', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(A_grounded,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'A_floating', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(A_floating,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'V_af', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(V_af,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'V_sle', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(V_sle,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'V_temp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(V_temp,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'A_temp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(A_temp,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'Q_s', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(Q_s,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'precip_tot', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(precip_tot,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'runoff_tot', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(runoff_tot,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'bmb_tot', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(bmb_tot,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'Q_b', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(Q_b,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'Q_temp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(Q_temp,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'calv_tot', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(calv_tot,sp), &
                         start=nc1cor), thisroutine )

#if (DISC>0)
call check( nf90_inq_varid(ncid, 'disc_lsc', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(disc_lsc,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'disc_ssc', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(disc_ssc,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'dT_glann', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(dT_glann,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'dT_sub', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(dT_sub,sp), &
                         start=nc1cor), thisroutine )
#endif

call check( nf90_inq_varid(ncid, 'dV_dt', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(dV_dt,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'mb_resid', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(mb_resid,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'mbp', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(mbp,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'H_max', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(H_max,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'H_t_max', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(H_t_max,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'zs_max', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(zs_max,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'vs_max', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(vs_max,sp), &
                         start=nc1cor), thisroutine )

call check( nf90_inq_varid(ncid, 'Tbh_max', ncv), thisroutine )
call check( nf90_put_var(ncid, ncv, real(Tbh_max,sp), &
                         start=nc1cor), thisroutine )

!  ------ Syncing NetCDF file (every 100th time)

n_sync = 100

if ( mod((counter-1), n_sync) == 0 ) &
   call check( nf90_sync(ncid),  thisroutine )

#endif

if (firstcall) firstcall = .false.

end subroutine output2

!-------------------------------------------------------------------------------
!> Writing of time-series data of the deep ice cores on file in ASCII format
!! (and optionally in NetCDF format).
!<------------------------------------------------------------------------------
subroutine output4(time, dxi, deta, delta_ts, glac_index, z_sl)

#if (NETCDF>1)
  use netcdf
  use nc_check_m
#endif

implicit none

real(dp), intent(in) :: time, dxi, deta, delta_ts, glac_index, z_sl

integer(i4b)                        :: i, j, n
real(sp), dimension(:), allocatable :: r_n_core
real(dp)                            :: time_val
real(dp), dimension(0:JMAX,0:IMAX)  :: field
real(dp), dimension(:), allocatable :: H_core, temp_b_core, &
                                       vx_b_core, vy_b_core, vh_b_core, &
                                       vx_s_core, vy_s_core, vh_s_core, &
                                       Rx_b_core, Ry_b_core, R_b_core, &
                                       bmb_core

#if (NETCDF>1)
integer(i4b), save :: ncid
integer(i4b)       :: ncd, ncv, nc1d, nc2d(2)
integer(i4b)       :: nc1cor(1), nc1cnt(1), nc2cor(2), nc2cnt(2)
integer(i4b)       :: n_sync
real(dp), save     :: time_add_offset_val
character(len= 16) :: ch_date, ch_time, ch_zone
character(len=256) :: filename, filename_with_path, buffer
logical, save      :: grads_nc_tweaks
#endif

integer(i4b), save :: counter   = 0
logical,      save :: firstcall = .true.

character(len=64), parameter :: thisroutine = 'output4'

counter = counter + 1

if (n_core >= 1) then

   allocate(r_n_core(n_core), H_core(n_core), temp_b_core(n_core), &
            vx_b_core(n_core), vy_b_core(n_core), vh_b_core(n_core), &
            vx_s_core(n_core), vy_s_core(n_core), vh_s_core(n_core), &
            Rx_b_core(n_core), Ry_b_core(n_core), R_b_core(n_core), &
            bmb_core(n_core))

!-------- Determination of ice-core data --------

   do n=1, n_core

!  ------Ice core number

      r_n_core(n) = real(n,sp)

!  ------ Ice thickness

      field = H_c + H_t
      call borehole(field, x_core(n), y_core(n), dxi, deta, 'grid', H_core(n))

!  ------ Basal velocity

      do i=0, IMAX
      do j=0, JMAX
         field(j,i) = vx_t(0,j,i)
      end do
      end do

      call borehole(field, x_core(n), y_core(n), dxi, deta, 'sg_x', vx_b_core(n))

      do i=0, IMAX
      do j=0, JMAX
         field(j,i) = vy_t(0,j,i)
      end do
      end do

      call borehole(field, x_core(n), y_core(n), dxi, deta, 'sg_y', vy_b_core(n))

      vh_b_core(n) = sqrt(vx_b_core(n)**2+vy_b_core(n)**2)

!  ------ Surface velocity

      do i=0, IMAX
      do j=0, JMAX
         field(j,i) = vx_c(KCMAX,j,i)
      end do
      end do

      call borehole(field, x_core(n), y_core(n), dxi, deta, 'sg_x', vx_s_core(n))

      do i=0, IMAX
      do j=0, JMAX
         field(j,i) = vy_c(KCMAX,j,i)
      end do
      end do

      call borehole(field, x_core(n), y_core(n), dxi, deta, 'sg_y', vy_s_core(n))

      vh_s_core(n) = sqrt(vx_s_core(n)**2+vy_s_core(n)**2)

!  ------ Basal temperature

      do i=0, IMAX
      do j=0, JMAX
         field(j,i) = temp_r(KRMAX,j,i)
      end do
      end do

      call borehole(field, x_core(n), y_core(n), dxi, deta, 'grid', temp_b_core(n))

!  ------ Basal frictional heating

      do i=0, IMAX
      do j=0, JMAX
         field(j,i) = vx_t(0,j,i)*txz_t(0,j,i)
      end do
      end do

      call borehole(field, x_core(n), y_core(n), dxi, deta, 'sg_x', Rx_b_core(n))

      do i=0, IMAX
      do j=0, JMAX
         field(j,i) = vy_t(0,j,i)*tyz_t(0,j,i)
      end do
      end do

      call borehole(field, x_core(n), y_core(n), dxi, deta, 'sg_y', Ry_b_core(n))

      R_b_core(n) = Rx_b_core(n) + Ry_b_core(n)

!  ------ Basal mass balance

      field = -(Q_bm+Q_tld)   ! positive for supply, negative for loss
      call borehole(field, x_core(n), y_core(n), dxi, deta, 'grid', bmb_core(n))

   end do

!  ------ Conversion

#if (!defined(OUT_TIMES) || OUT_TIMES==1)
   time_val = time *sec_to_year   ! s -> a
#elif (OUT_TIMES==2)
   time_val = (time+year_zero) *sec_to_year   ! s -> a
#else
   errormsg = ' >>> output4: OUT_TIMES must be either 1 or 2!'
   call error(errormsg)
#endif

   vh_b_core = vh_b_core *YEAR_SEC   ! m/s -> m/a
   vh_s_core = vh_s_core *YEAR_SEC   ! m/s -> m/a

   bmb_core  = bmb_core  *YEAR_SEC   ! m ice equiv./s -> m ice equiv./a

!-------- Writing of data on file --------

   if (forcing_flag == 1) then
      write(unit=14, fmt='(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   else if (forcing_flag == 2) then
      write(unit=14, fmt='(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   else if (forcing_flag == 3) then
      write(unit=14, fmt='(1pe13.6,2(1pe13.4))') time_val, no_value_pos_1, z_sl
   end if

   n=1
   write(unit=14, fmt='(13x,1pe13.4)', advance='no') H_core(n)
   do n=2, n_core-1
      write(unit=14, fmt='(1pe13.4)', advance='no') H_core(n)
   end do
   n=n_core
   write(unit=14, fmt='(1pe13.4)') H_core(n)

   n=1
   write(unit=14, fmt='(13x,1pe13.4)', advance='no') vh_s_core(n)
   do n=2, n_core-1
      write(unit=14, fmt='(1pe13.4)', advance='no') vh_s_core(n)
   end do
   n=n_core
   write(unit=14, fmt='(1pe13.4)') vh_s_core(n)

   n=1
   write(unit=14, fmt='(13x,1pe13.4)', advance='no') temp_b_core(n)
   do n=2, n_core-1
      write(unit=14, fmt='(1pe13.4)', advance='no') temp_b_core(n)
   end do
   n=n_core
   write(unit=14, fmt='(1pe13.4,/)') temp_b_core(n)

!-------- Extended time-series file in NetCDF format --------

#if (NETCDF>1)

   if (firstcall) then

!  ------ Open NetCDF file

      filename           = trim(RUNNAME)//'_core.nc'
      filename_with_path = trim(OUTPATH)//'/'//trim(filename)

      call check( nf90_create(trim(filename_with_path), NF90_NOCLOBBER, ncid), &
                  thisroutine )

      ncid_core = ncid

!  ------ Global attributes

      buffer = 'Time-series output for the deep ice cores of simulation '// &
               trim(RUNNAME)
      call check( nf90_put_att(ncid, NF90_GLOBAL, 'title', trim(buffer)), &
                  thisroutine )

      buffer = 'Institute of Low Temperature Science, Hokkaido University, '// &
               'Sapporo, Japan'
      call check( nf90_put_att(ncid, NF90_GLOBAL, 'institution', &
                               trim(buffer)), &
                  thisroutine )

      buffer = 'SICOPOLIS Version '//VERSION
      call check( nf90_put_att(ncid, NF90_GLOBAL, 'source', trim(buffer)), &
                  thisroutine )

      call date_and_time(ch_date, ch_time, ch_zone)
      buffer = ch_date(1:4)//'-'//ch_date(5:6)//'-'//ch_date(7:8)//' '// &
               ch_time(1:2)//':'//ch_time(3:4)//':'//ch_time(5:6)//' '// &
               ch_zone(1:3)//':'//ch_zone(4:5)//' - Data produced'
      call check( nf90_put_att(ncid, NF90_GLOBAL, 'history', trim(buffer)), &
                  thisroutine )

      buffer = 'http://www.sicopolis.net/'
      call check( nf90_put_att(ncid, NF90_GLOBAL, 'references', trim(buffer)), &
                  thisroutine )

!  ------ Definition of the dimensions

      call set_grads_nc_tweaks(grads_nc_tweaks)

      if (grads_nc_tweaks) then
         call check( nf90_def_dim(ncid, 'x', 1, ncd), thisroutine )
         call check( nf90_def_dim(ncid, 'y', 1, ncd), thisroutine )
      end if

      call check( nf90_def_dim(ncid, 'n', n_core,  ncd), thisroutine )
      call check( nf90_def_dim(ncid, 't', NF90_UNLIMITED, ncd), thisroutine )

!  ------ Definition of the variables

      if (grads_nc_tweaks) then

!    ---- x

         call check( nf90_inq_dimid(ncid, 'x', nc1d), thisroutine )
         call check( nf90_def_var(ncid, 'x', NF90_FLOAT, nc1d, ncv), &
                     thisroutine )
         buffer = 'm'
         call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                     thisroutine )
         buffer = 'x'
         call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                     thisroutine )
         buffer = 'Dummy x-coordinate for one point'
         call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                     thisroutine )
         call check( nf90_put_att(ncid, ncv, 'axis', 'x'), thisroutine )

!    ---- y

         call check( nf90_inq_dimid(ncid, 'y', nc1d), thisroutine )
         call check( nf90_def_var(ncid, 'y', NF90_FLOAT, nc1d, ncv), &
                     thisroutine )
         buffer = 'm'
         call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                     thisroutine )
         buffer = 'y'
         call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                     thisroutine )
         buffer = 'Dummy y-coordinate for one point'
         call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                     thisroutine )
         call check( nf90_put_att(ncid, ncv, 'axis', 'y'), thisroutine )

      end if

!    ---- Time

      call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 't', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = 'a'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'time'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Time'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )
      call check( nf90_put_att(ncid, ncv, 'axis', 't'), thisroutine )

      if (grads_nc_tweaks) then

!    ---- Time offset

         call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
         call check( nf90_def_var(ncid, 't_add_offset', &
                                  NF90_FLOAT, nc1d, ncv), &
                     thisroutine )
         buffer = 'a'
         call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                     thisroutine )
         buffer = 'time_add_offset'
         call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                     thisroutine )
         buffer = 'Time offset'
         call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                     thisroutine )

         time_add_offset_val = min(time_val, 0.0_dp)

      end if

!    ---- Ice core number

      call check( nf90_inq_dimid(ncid, 'n', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 'n', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = '1'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'ice_core_number'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Ice core number'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )
      buffer = trim(ch_core(1))
      do n=2, n_core; buffer = trim(buffer)//', '//trim(ch_core(n)); end do
      call check( nf90_put_att(ncid, ncv, 'ice_core_names', trim(buffer)), &
                  thisroutine )
      call check( nf90_put_att(ncid, ncv, 'axis', 'n'), thisroutine )

      if (forcing_flag == 1) then

!    ---- delta_ts

         call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
         call check( nf90_def_var(ncid, 'delta_ts', NF90_FLOAT, nc1d, ncv), &
                     thisroutine )
         buffer = 'degC'
         call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                     thisroutine )
         buffer = 'surface_temperature_anomaly'
         call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                     thisroutine )
         buffer = 'Surface temperature anomaly'
         call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                     thisroutine )

      else if ((forcing_flag == 2).or.(forcing_flag == 3)) then

!    ---- glac_index

         call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
         call check( nf90_def_var(ncid, 'glac_index', NF90_FLOAT, nc1d, ncv), &
                     thisroutine )
         buffer = '1'
         call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                     thisroutine )
         buffer = 'glacial_index'
         call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                     thisroutine )
         buffer = 'Glacial index'
         call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                     thisroutine )

      end if

!    ---- z_sl

      call check( nf90_inq_dimid(ncid, 't', nc1d), thisroutine )
      call check( nf90_def_var(ncid, 'z_sl', NF90_FLOAT, nc1d, ncv), &
                  thisroutine )
      buffer = 'm'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'global_average_sea_level_change'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Sea level'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

!    ---- H_core

      call check( nf90_inq_dimid(ncid, 'n', nc2d(1)), thisroutine )
      call check( nf90_inq_dimid(ncid, 't', nc2d(2)), thisroutine )
      call check( nf90_def_var(ncid, 'H_core', NF90_FLOAT, nc2d, ncv), &
                  thisroutine )
      buffer = 'm'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'land_ice_thickness'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Ice thickness'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

!    ---- vh_b_core

      call check( nf90_inq_dimid(ncid, 'n', nc2d(1)), thisroutine )
      call check( nf90_inq_dimid(ncid, 't', nc2d(2)), thisroutine )
      call check( nf90_def_var(ncid, 'vh_b_core', NF90_FLOAT, nc2d, ncv), &
                  thisroutine )
      buffer = 'm a-1'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'land_ice_basal_horizontal_velocity'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Horizontal velocity at the ice base'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

!    ---- vh_s_core

      call check( nf90_inq_dimid(ncid, 'n', nc2d(1)), thisroutine )
      call check( nf90_inq_dimid(ncid, 't', nc2d(2)), thisroutine )
      call check( nf90_def_var(ncid, 'vh_s_core', NF90_FLOAT, nc2d, ncv), &
                  thisroutine )
      buffer = 'm a-1'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'land_ice_surface_horizontal_velocity'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Horizontal velocity at the ice surface'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

!    ---- temp_b_core

      call check( nf90_inq_dimid(ncid, 'n', nc2d(1)), thisroutine )
      call check( nf90_inq_dimid(ncid, 't', nc2d(2)), thisroutine )
      call check( nf90_def_var(ncid, 'temp_b_core', NF90_FLOAT, nc2d, ncv), &
                  thisroutine )
      buffer = 'degC'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'basal_temperature'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Temperature at the ice base'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

!    ---- R_b_core

      call check( nf90_inq_dimid(ncid, 'n', nc2d(1)), thisroutine )
      call check( nf90_inq_dimid(ncid, 't', nc2d(2)), thisroutine )
      call check( nf90_def_var(ncid, 'R_b_core', NF90_FLOAT, nc2d, ncv), &
                  thisroutine )
      buffer = 'W m-2'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'basal_frictional_heating'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Basal frictional heating'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

!    ---- bmb_core

      call check( nf90_inq_dimid(ncid, 'n', nc2d(1)), thisroutine )
      call check( nf90_inq_dimid(ncid, 't', nc2d(2)), thisroutine )
      call check( nf90_def_var(ncid, 'bmb_core', NF90_FLOAT, nc2d, ncv), &
                  thisroutine )
      buffer = 'm ice equiv. a-1'
      call check( nf90_put_att(ncid, ncv, 'units', trim(buffer)), &
                  thisroutine )
      buffer = 'land_ice_basal_mass_balance'
      call check( nf90_put_att(ncid, ncv, 'standard_name', trim(buffer)), &
                  thisroutine )
      buffer = 'Basal mass balance'
      call check( nf90_put_att(ncid, ncv, 'long_name', trim(buffer)), &
                  thisroutine )

!    ---- End of the definitions

      call check( nf90_enddef(ncid), thisroutine )

   end if

!  ------ Writing of data on NetCDF file

   if (firstcall.and.grads_nc_tweaks) then

      nc1cor = (/ 1 /)

      call check( nf90_inq_varid(ncid, 'x', ncv), thisroutine )
      call check( nf90_put_var(ncid, ncv, 0.0_sp, start=nc1cor), thisroutine )

      call check( nf90_inq_varid(ncid, 'y', ncv), thisroutine )
      call check( nf90_put_var(ncid, ncv, 0.0_sp, start=nc1cor), thisroutine )

      call check( nf90_sync(ncid), thisroutine )

   end if

   if (firstcall) then

      nc1cor = (/ 1 /)

      call check( nf90_inq_varid(ncid, 'n', ncv), thisroutine )
      call check( nf90_put_var(ncid, ncv, r_n_core, &
                               start=nc1cor), thisroutine )

   end if

   nc1cor(1) = counter
   ! nc1cnt(1) = 1   ! (not needed, and causes troubles for whatever reason)

   nc2cor(1) = 1
   nc2cor(2) = counter
   nc2cnt(1) = n_core
   nc2cnt(2) = 1

   call check( nf90_inq_varid(ncid, 't', ncv), thisroutine )

   if (.not.grads_nc_tweaks) then
      call check( nf90_put_var(ncid, ncv, real(time_val,sp), &
                               start=nc1cor), thisroutine )
   else
      call check( nf90_put_var(ncid, ncv, &
                               real(time_val-time_add_offset_val,sp), &
                               start=nc1cor), thisroutine )
                            ! this makes sure that all times are >=0
                            ! (GrADS doesn't like negative numbers)
      call check( nf90_inq_varid(ncid, 't_add_offset', ncv), thisroutine )
      call check( nf90_put_var(ncid, ncv, &
                               real(time_add_offset_val,sp), &
                               start=nc1cor), thisroutine )
   end if

   if (forcing_flag == 1) then

      call check( nf90_inq_varid(ncid, 'delta_ts', ncv), thisroutine )
      call check( nf90_put_var(ncid, ncv, real(delta_ts,sp), &
                               start=nc1cor), thisroutine )

   else if (forcing_flag == 2) then

      call check( nf90_inq_varid(ncid, 'glac_index', ncv), thisroutine )
      call check( nf90_put_var(ncid, ncv, real(glac_index,sp), &
                               start=nc1cor), thisroutine )

   else if (forcing_flag == 3) then

      call check( nf90_inq_varid(ncid, 'glac_index', ncv), thisroutine )
      call check( nf90_put_var(ncid, ncv, real(no_value_neg_2,sp), &
                               start=nc1cor), thisroutine )

   end if

   call check( nf90_inq_varid(ncid, 'z_sl', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(z_sl,sp), &
                            start=nc1cor), thisroutine )

   call check( nf90_inq_varid(ncid, 'H_core', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(H_core,sp), &
                            start=nc2cor, count=nc2cnt), thisroutine )

   call check( nf90_inq_varid(ncid, 'vh_b_core', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(vh_b_core,sp), &
                            start=nc2cor, count=nc2cnt), thisroutine )

   call check( nf90_inq_varid(ncid, 'vh_s_core', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(vh_s_core,sp), &
                            start=nc2cor, count=nc2cnt), thisroutine )

   call check( nf90_inq_varid(ncid, 'temp_b_core', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(temp_b_core,sp), &
                            start=nc2cor, count=nc2cnt), thisroutine )

   call check( nf90_inq_varid(ncid, 'R_b_core', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(R_b_core,sp), &
                            start=nc2cor, count=nc2cnt), thisroutine )

   call check( nf90_inq_varid(ncid, 'bmb_core', ncv), thisroutine )
   call check( nf90_put_var(ncid, ncv, real(bmb_core,sp), &
                            start=nc2cor, count=nc2cnt), thisroutine )

!  ------ Syncing NetCDF file (every 100th time)

   n_sync = 100

   if ( mod((counter-1), n_sync) == 0 ) &
      call check( nf90_sync(ncid),  thisroutine )

#endif

   deallocate(r_n_core, H_core, &
              vx_b_core, vy_b_core, vh_b_core, &
              vx_s_core, vy_s_core, vh_s_core, &
              temp_b_core, &
              Rx_b_core, Ry_b_core, R_b_core, &
              bmb_core)

!!! else   ! (n_core == 0 -> do nothing)
!!!
!!!    continue

end if

if (firstcall) firstcall = .false.

end subroutine output4

#if (defined(ASF))

!-------------------------------------------------------------------------------
!> Writing of time-series data for all defined surface points on file
!! in ASCII format. Modification of Tolly's output7 by Thorben Dunse.
!<------------------------------------------------------------------------------
subroutine output5(time, dxi, deta, delta_ts, glac_index, z_sl)

implicit none

real(dp), intent(in) :: time, dxi, deta, delta_ts, glac_index, z_sl

integer(i4b) :: n, k
real(dp) :: time_val
real(dp), dimension(0:JMAX,0:IMAX) :: field
real(dp), dimension(:), allocatable :: zl_surf, zs_surf, &
                  accum_surf, as_perp_surf, &
		  snowfall_surf, rainfall_surf, runoff_surf, &
		  vx_surf, vy_surf, vz_surf, &
                  vx_base, vy_base, vz_base, &
		  temp_base_pmp

allocate(zl_surf(n_surf), zs_surf(n_surf), &
        accum_surf(n_surf), &
	as_perp_surf(n_surf), snowfall_surf(n_surf), &
	rainfall_surf(n_surf), runoff_surf(n_surf), &
	vx_surf(n_surf), vy_surf(n_surf), vz_surf(n_surf), &
	vx_base(n_surf), vy_base(n_surf), vz_base(n_surf), &
 	temp_base_pmp(n_surf))

!-------- Determination of ice-core data --------

do n=1, n_surf

!  ------ Bedrock elevation

   field = zl
   call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', zl_surf(n))

!  ------ Surface elevation

   field = zs
   call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', zs_surf(n))


!  ------ Accumulation

   field = accum
   call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', accum_surf(n))

!  ------ Surface mass balance

   field = as_perp
   call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', as_perp_surf(n))

!  ------ Snowfall

   field = snowfall
   call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', snowfall_surf(n))

!  ------ Rainfall

   field = rainfall
   call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', rainfall_surf(n))

!  ------ Runoff

   field = runoff
   call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', runoff_surf(n))

! ------ Surface velocities

      field = vx_s_g
      call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', vx_surf(n))

      field = vy_s_g
      call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', vy_surf(n))

      field = vz_s
      call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', vz_surf(n))

! ------ Basal velocities

      field = vx_b_g
      call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', vx_base(n))

      field = vy_b_g
      call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', vy_base(n))

      field = vz_b
      call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', vz_base(n))

! ------ Basal temperature relative to pressure melting point

      field = temph_b
      call borehole(field, x_surf(n), y_surf(n), dxi, deta, &
	'grid', temp_base_pmp(n))


end do

!  ------ Conversion

#if (!defined(OUT_TIMES) || OUT_TIMES==1)
time_val = time *sec_to_year   ! s -> a
#elif (OUT_TIMES==2)
time_val = (time+year_zero) *sec_to_year   ! s -> a
#else
errormsg = ' >>> output5: OUT_TIMES must be either 1 or 2!'
call error(errormsg)
#endif

do n=1, n_surf
   accum_surf(n)    = accum_surf(n)    *YEAR_SEC   ! m/s -> m/a
   as_perp_surf(n)  = as_perp_surf(n)  *YEAR_SEC   ! m/s -> m/a
   snowfall_surf(n) = snowfall_surf(n) *YEAR_SEC   ! m/s -> m/a
   rainfall_surf(n) = rainfall_surf(n) *YEAR_SEC   ! m/s -> m/a
   runoff_surf(n)   = runoff_surf(n)   *YEAR_SEC   ! m/s -> m/a
   vx_surf(n)       = vx_surf(n)       *YEAR_SEC   ! m/s -> m/a
   vy_surf(n)       = vy_surf(n)       *YEAR_SEC   ! m/s -> m/a
   vz_surf(n)       = vz_surf(n)       *YEAR_SEC   ! m/s -> m/a
   vx_base(n)       = vx_base(n)       *YEAR_SEC   ! m/s -> m/a
   vy_base(n)       = vy_base(n)       *YEAR_SEC   ! m/s -> m/a
   vz_base(n)       = vz_base(n)       *YEAR_SEC   ! m/s -> m/a
end do

!-------- Writing of data on file --------

if (forcing_flag == 1) then
   write(41,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(42,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(43,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(44,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(45,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(46,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(47,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(48,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(49,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(50,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(51,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(52,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(53,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
   write(54,'(1pe13.6,2(1pe13.4))') time_val, delta_ts, z_sl
else if (forcing_flag == 2) then
   write(41,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(42,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(43,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(44,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(45,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(46,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(47,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(48,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(49,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(50,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(51,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(52,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(53,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
   write(54,'(1pe13.6,2(1pe13.4))') time_val, glac_index, z_sl
end if

do n=1, n_surf-1
   write(41,'(1pe13.4)',advance='no') zl_surf(n)
   write(42,'(1pe13.4)',advance='no') zs_surf(n)
   write(43,'(1pe13.4)',advance='no') accum_surf(n)
   write(44,'(1pe13.4)',advance='no') as_perp_surf(n)
   write(45,'(1pe13.4)',advance='no') snowfall_surf(n)
   write(46,'(1pe13.4)',advance='no') rainfall_surf(n)
   write(47,'(1pe13.4)',advance='no') runoff_surf(n)
   write(48,'(1pe13.4)',advance='no') vx_surf(n)
   write(49,'(1pe13.4)',advance='no') vy_surf(n)
   write(50,'(1pe13.4)',advance='no') vz_surf(n)
   write(51,'(1pe13.4)',advance='no') vx_base(n)
   write(52,'(1pe13.4)',advance='no') vy_base(n)
   write(53,'(1pe13.4)',advance='no') vz_base(n)
   write(54,'(1pe13.4)',advance='no') temp_base_pmp(n) 
end do

n=n_surf
   write(41,'(1pe13.4)') zl_surf(n)
   write(42,'(1pe13.4)') zs_surf(n)
   write(43,'(1pe13.4)') accum_surf(n)
   write(44,'(1pe13.4)') as_perp_surf(n)
   write(45,'(1pe13.4)') snowfall_surf(n)
   write(46,'(1pe13.4)') rainfall_surf(n)
   write(47,'(1pe13.4)') runoff_surf(n)
   write(48,'(1pe13.4)') vx_surf(n)
   write(49,'(1pe13.4)') vy_surf(n)
   write(50,'(1pe13.4)') vz_surf(n)
   write(51,'(1pe13.4)') vx_base(n)
   write(52,'(1pe13.4)') vy_base(n)
   write(53,'(1pe13.4)') vz_base(n)
   write(54,'(1pe13.4)') temp_base_pmp(n) 

deallocate(zl_surf, zs_surf, accum_surf, as_perp_surf, &
           snowfall_surf, rainfall_surf, runoff_surf, &
	   vx_surf, vy_surf, vz_surf, &
	   vx_base, vy_base, vz_base,temp_base_pmp)

end subroutine output5

#endif   /* (defined(ASF)) */

!-------------------------------------------------------------------------------
!> Computation of an arbitrary field quantity for a given borehole
!! position x_pos, y_pos by weighed averaging of the corresponding
!! gridded 2-d field.
!<------------------------------------------------------------------------------
  subroutine borehole(field, x_pos, y_pos, dxi, deta, ch_grid, field_val)

#if !defined(ALLOW_OPENAD) /* Normal */
  use sico_maths_m, only : bilinint
#else /* OpenAD */
  use sico_maths_m
#endif /* Normal vs. OpenAD */

  implicit none

  real(dp), dimension(0:JMAX,0:IMAX), intent(in) :: field
  real(dp), intent(in) :: x_pos, y_pos, dxi, deta
  character (len=*), intent(in) :: ch_grid

  real(dp), intent(out) :: field_val

  integer(i4b) :: i1, i2, j1, j2
  real(dp) :: real_i, real_j

!-------- Neighbour points --------

  real_i = (x_pos-xi(0)) /dxi
  real_j = (y_pos-eta(0))/deta

  if (ch_grid=='sg_x') real_i = real_i - 0.5_dp
  if (ch_grid=='sg_y') real_j = real_j - 0.5_dp

  if (real_i < 0.5_dp*real(IMAX,dp)) then
     i1 = floor(real_i)
     i2 = i1 + 1
  else
     i2 = ceiling(real_i)
     i1 = i2 - 1
  end if

  if (real_j < 0.5_dp*real(JMAX,dp)) then
     j1 = floor(real_j)
     j2 = j1 + 1
  else
     j2 = ceiling(real_j)
     j1 = j2 - 1
  end if

  if (ch_grid=='grid') then
                    ! field(j,i) defined on grid points

     if ((i1 < 0).or.(j1 < 0).or.(i2 > IMAX).or.(j2 > JMAX)) then
        errormsg = ' >>> borehole: Borehole position out of domain!'
        call error(errormsg)
     end if

  else if (ch_grid=='sg_x') then
                    ! field(j,i) defined on staggered grid in x direction

     if ((i1 < 0).or.(j1 < 0).or.(i2 > IMAX-1).or.(j2 > JMAX)) then
        errormsg = ' >>> borehole: Borehole position out of domain!'
        call error(errormsg)
     end if

  else if (ch_grid=='sg_y') then
                    ! field(j,i) defined on staggered grid in y direction

     if ((i1 < 0).or.(j1 < 0).or.(i2 > IMAX).or.(j2 > JMAX-1)) then
        errormsg = ' >>> borehole: Borehole position out of domain!'
        call error(errormsg)
     end if

  else

     errormsg = ' >>> borehole: Parameter ch_grid has undefined value!'
     call error(errormsg)

  end if

!-------- Weighing of the four adjacent grid values --------

#if !defined(ALLOW_OPENAD) /* Normal */
  field_val = bilinint(real(i1,dp), real(i2,dp), real(j1,dp), real(j2,dp), &
                       field(j1,i1), field(j2,i1), field(j1,i2), field(j2,i2), &
                       real_i, real_j)
#else /* OpenAD */
  call bilinint(real(i1,dp), real(i2,dp), real(j1,dp), real(j2,dp), &
                       field(j1,i1), field(j2,i1), field(j1,i2), field(j2,i2), &
                       real_i, real_j, field_val)
#endif /* Normal vs. OpenAD */

  end subroutine borehole

!-------------------------------------------------------------------------------
!> Set the value of the auxiliary variable grads_nc_tweaks.
!<------------------------------------------------------------------------------
  subroutine set_grads_nc_tweaks(grads_nc_tweaks)

  implicit none

  logical, intent(out) :: grads_nc_tweaks

  character(len=16) :: ch_value

#if (NETCDF>1)

  grads_nc_tweaks = .false.

!-------- Try environment variable --------

  call get_environment_variable('SICO_GRADS_NC_TWEAKS', ch_value)

  if ( (trim(ch_value)=='true') &
       .or.(trim(ch_value)=='True').or.(trim(ch_value)=='TRUE') ) &
     grads_nc_tweaks = .true.

  if ( (trim(ch_value)=='yes') &   ! obsolete, but still supported
       .or.(trim(ch_value)=='Yes').or.(trim(ch_value)=='YES') &
       .or.(trim(ch_value)=='y').or.(trim(ch_value)=='Y') ) &
     grads_nc_tweaks = .true.

!-------- Try preprocessor switch --------

#if (defined(GRADS_NC_TWEAKS))
#if (GRADS_NC_TWEAKS==1)
  grads_nc_tweaks = .true.
#else
  grads_nc_tweaks = .false.
#endif
#endif

#else   /* NetCDF not used */
  grads_nc_tweaks = .false.   ! dummy value
#endif

  end subroutine set_grads_nc_tweaks

!-------------------------------------------------------------------------------

end module output_m
!
